"""Finviz Equity Screener Model."""

# pylint: disable=unused-argument,too-many-statements,too-many-branches,too-many-locals

from typing import Any, Dict, List, Literal, Optional, Union
from warnings import warn

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_screener import (
    EquityScreenerData,
    EquityScreenerQueryParams,
)
from openbb_core.provider.utils.descriptions import (
    DATA_DESCRIPTIONS,
    QUERY_DESCRIPTIONS,
)
from openbb_core.provider.utils.errors import EmptyDataError
from openbb_finviz.utils.screener_helper import (
    EXCHANGE_MAP,
    INDEX_MAP,
    INDUSTRY_MAP,
    MARKET_CAP_MAP,
    RECOMMENDATION_MAP,
    SECTOR_MAP,
    SIGNALS,
    SIGNALS_DESC_STR,
    MarketCap,
    Recommendation,
    Sectors,
)
from pydantic import Field, field_validator


class FinvizEquityScreenerQueryParams(EquityScreenerQueryParams):
    """Finviz Equity Screener Query Params."""

    __json_schema_extra__ = {
        "metric": {
            "choices": [
                "overview",
                "valuation",
                "financial",
                "ownership",
                "performance",
                "technical",
            ],
        },
        "signal": {"choices": SIGNALS},
        "exchange": {"choices": list(EXCHANGE_MAP)},
        "index": {"choices": list(INDEX_MAP)},
        "sector": {"choices": list(SECTOR_MAP)},
        "industry": {"choices": list(INDUSTRY_MAP)},
        "market_cap": {"choices": list(MARKET_CAP_MAP)},
        "recommendation": {"choices": list(RECOMMENDATION_MAP)},
    }

    metric: Literal[
        "overview",
        "valuation",
        "financial",
        "ownership",
        "performance",
        "technical",
    ] = Field(
        default="overview",
        description="The data group to return, default is 'overview'.",
    )
    exchange: Literal["all", "amex", "nasdaq", "nyse"] = Field(
        default="all", description="Filter by exchange."
    )
    index: Literal["all", "dow", "nasdaq", "sp500", "russell"] = Field(
        default="all",
        description="Filter by index.",
    )
    sector: Sectors = Field(
        default="all",
        description="Filter by sector.",
    )
    industry: Optional[str] = Field(
        default="all",
        description="Filter by industry.",
    )
    mktcap: MarketCap = Field(
        default="all",
        description="Filter by market cap."
        + "\n    Mega - > 200B"
        + "\n    Large - 10B - 200B"
        + "\n    Mid - 2B - 10B"
        + "\n    Small - 300M - 2B"
        + "\n    Micro - 50M - 300M"
        + "\n    Nano - < 50M",
    )
    recommendation: Recommendation = Field(
        default="all", description="Filter by analyst recommendation."
    )
    signal: Optional[str] = Field(
        default=None,
        description="The Finviz screener signal to use."
        + " When no parameters are provided, the screener defaults to 'top_gainers'."
        + f" Available signals are:{SIGNALS_DESC_STR}",
    )
    preset: Optional[str] = Field(
        default=None,
        description="A configured preset file to use for the query."
        + " This overrides all other query parameters except 'metric', and 'limit'."
        + " Presets (.ini text files) can be created and modified in the '~/OpenBBUserData/finviz/presets' directory."
        + " If the path does not exist, it will be created and populated with the default presets on the first run."
        + " Refer to the file, 'screener_template.ini', for the format and options."
        + "\n\nNote: Syntax of parameters in preset files must follow the template file exactly "
        + " - i.e, Analyst Recom. = Strong Buy (1)",
    )
    filters_dict: Optional[Union[Dict, str]] = Field(
        default=None,
        kw_only=True,
        description="A formatted dictionary, or serialized JSON string, of additional filters to apply to the query."
        + " This parameter can be used as an alternative to preset files, and is ignored when a preset is supplied."
        + " Invalid entries will raise an error. Syntax should follow the 'screener_template.ini' file.",
    )
    limit: Optional[int] = Field(
        default=None,
        description=QUERY_DESCRIPTIONS.get("limit", ""),
    )

    @field_validator("signal", mode="before", check_fields=False)
    @classmethod
    def validate_signal(cls, v):
        """Validate the signal."""
        if v is not None and v not in SIGNALS:
            raise OpenBBError(
                f"Invalid signal '{v}'. Available signals are:\n{SIGNALS_DESC_STR}"
            )
        return v if v else None

    @field_validator("industry", mode="before", check_fields=False)
    @classmethod
    def validate_industry(cls, v):
        """Validate the industry."""
        if v is not None and v not in INDUSTRY_MAP:
            raise OpenBBError(
                f"Invalid industry '{v}'. Available industries are:\n{', '.join(INDUSTRY_MAP)}"
            )
        return v if v else None

    @field_validator("preset", mode="before", check_fields=False)
    @classmethod
    def validate_preset(cls, v):
        """Check to reject running template file."""
        if v is not None and v == "screener_template":
            raise OpenBBError(
                f"Invalid preset '{v}'. Please rename the file to use as a preset."
            )
        return v if v else None

    @field_validator("filters_dict", mode="before", check_fields=False)
    @classmethod
    def validate_filters_dict(cls, v):
        """Validate the filters_dict."""
        if isinstance(v, str):
            # pylint: disable=import-outside-toplevel
            import json

            try:
                v = json.loads(v)
            except json.JSONDecodeError as e:
                raise OpenBBError(f"Invalid JSON format for 'filters_dict': {e}") from e
        if v is not None and not isinstance(v, dict):
            raise OpenBBError(
                "Invalid 'filters_dict' format. Must be a dictionary or serialized JSON string."
            )

        return v


class FinvizEquityScreenerData(EquityScreenerData):
    """Finviz Equity Screener Data. Actual returned data varies by the 'metric' parameter."""

    __alias_dict__ = {
        "symbol": "Ticker",
        "name": "Company",
        "earnings_date": "Earnings",
        "sector": "Sector",
        "industry": "Industry",
        "country": "Country",
        "shares_outstanding": "Outstanding",
        "shares_float": "Float",
        "short_interest": "Float Short",
        "short_ratio": "Short Ratio",
        "insider_ownership": "Insider Own",
        "insider_ownership_change": "Insider Trans",
        "institutional_ownership": "Inst Own",
        "institutional_ownership_change": "Inst Trans",
        "analyst_recommendation": "Recom",
        "beta": "Beta",
        "market_cap": "Market Cap",
        "price": "Price",
        "change_percent": "Change",
        "change_from_open": "from Open",
        "gap": "Gap",
        "year_high_percent": "52W High",
        "year_low_percent": "52W Low",
        "sma20_percent": "SMA20",
        "sma50_percent": "SMA50",
        "sma200_percent": "SMA200",
        "rsi": "RSI",
        "volume": "Volume",
        "volume_avg": "Avg Volume",
        "volume_relative": "Rel Volume",
        "average_true_range": "ATR",
        "price_change_1w": "Perf Week",
        "price_change_1m": "Perf Month",
        "price_change_3m": "Perf Quart",
        "price_change_6m": "Perf Half",
        "price_change_1y": "Perf Year",
        "price_change_ytd": "Perf YTD",
        "volatility_1w": "Volatility W",
        "volatility_1m": "Volatility M",
        "price_to_earnings": "P/E",
        "forward_pe": "Fwd P/E",
        "peg_ratio": "PEG",
        "price_to_sales": "P/S",
        "price_to_book": "P/B",
        "price_to_cash": "P/C",
        "price_to_free_cash_flow": "P/FCF",
        "eps_growth_past_1y": "EPS this Y",
        "eps_growth_next_1y": "EPS next Y",
        "eps_growth_past_5y": "EPS past 5Y",
        "eps_growth_next_5y": "EPS next 5Y",
        "sales_growth_past_5y": "Sales past 5Y",
        "dividend_yield": "Dividend",
        "return_on_assets": "ROA",
        "return_on_equity": "ROE",
        "return_on_investment": "ROI",
        "current_ratio": "Curr R",
        "quick_ratio": "Quick R",
        "long_term_debt_to_equity": "LTDebt/Eq",
        "debt_to_equity": "Debt/Eq",
        "gross_margin": "Gross M",
        "operating_margin": "Oper M",
        "profit_margin": "Profit M",
    }

    earnings_date: Optional[str] = Field(
        default=None,
        description="Earnings date, where 'a' and 'b' mean after and before market close, respectively.",
    )
    country: Optional[str] = Field(
        default=None,
        description="Country of the company.",
    )
    sector: Optional[str] = Field(
        default=None,
        description="Sector of the company.",
    )
    industry: Optional[str] = Field(
        default=None,
        description="Industry of the company.",
    )
    beta: Optional[float] = Field(
        default=None,
        description="Beta of the stock.",
    )
    analyst_recommendation: Optional[float] = Field(
        default=None,
        description="Analyst's mean recommendation. (1=Buy 5=Sell).",
    )
    market_cap: Optional[float] = Field(
        default=None,
        description="Market capitalization of the company.",
        json_schema_extra={"x-unit_measurement": "currency"},
    )
    price: Optional[float] = Field(
        default=None,
        description="Price of a share.",
        json_schema_extra={"x-unit_measurement": "currency"},
    )
    change_percent: Optional[float] = Field(
        default=None,
        description="Price change percentage.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    change_from_open: Optional[float] = Field(
        default=None,
        description="Price change percentage, from the opening price.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    gap: Optional[float] = Field(
        default=None,
        description="Price gap percentage, from the previous close.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    volume: Optional[Union[int, float]] = Field(
        default=None,
        description=DATA_DESCRIPTIONS.get("volume", ""),
    )
    volume_avg: Optional[Union[int, float]] = Field(
        default=None,
        description="3-month average daily volume.",
    )
    volume_relative: Optional[float] = Field(
        default=None,
        description="Current volume relative to the average.",
    )
    average_true_range: Optional[float] = Field(
        default=None,
        description="Average true range (14).",
        json_schema_extra={"x-unit_measurement:": "currency"},
    )
    price_change_1w: Optional[float] = Field(
        default=None,
        description="One-week price return.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    price_change_1m: Optional[float] = Field(
        default=None,
        description="One-month price return.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    price_change_3m: Optional[float] = Field(
        default=None,
        description="Three-month price return.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    price_change_6m: Optional[float] = Field(
        default=None,
        description="Six-month price return.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    price_change_1y: Optional[float] = Field(
        default=None,
        description="One-year price return.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    price_change_ytd: Optional[float] = Field(
        default=None,
        description="Year-to-date price return.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    volatility_1w: Optional[float] = Field(
        default=None,
        description="One-week volatility.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    volatility_1m: Optional[float] = Field(
        default=None,
        description="One-month volatility.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    year_high_percent: Optional[float] = Field(
        default=None,
        description="Percent difference from current price to the 52-week high.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    year_low_percent: Optional[float] = Field(
        default=None,
        description="Percent difference from current price to the 52-week low.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    sma20_percent: Optional[float] = Field(
        default=None,
        description="Percent difference from current price to the 20-day simple moving average.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    sma50_percent: Optional[float] = Field(
        default=None,
        description="Percent difference from current price to the 50-day simple moving average.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    sma200_percent: Optional[float] = Field(
        default=None,
        description="Percent difference from current price to the 200-day simple moving average.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    rsi: Optional[float] = Field(
        default=None,
        description="Relative strength index (14).",
    )
    shares_outstanding: Optional[Union[int, float]] = Field(
        default=None,
        description="Number of shares outstanding.",
    )
    shares_float: Optional[Union[int, float]] = Field(
        default=None,
        description="Number of shares available to trade.",
    )
    short_interest: Optional[float] = Field(
        default=None,
        description="Percent of float reported as short.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    short_ratio: Optional[float] = Field(
        default=None,
        description="Short interest ratio",
    )
    insider_ownership: Optional[float] = Field(
        default=None,
        description="Insider ownership as a percentage.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    insider_ownership_change: Optional[float] = Field(
        default=None,
        description="6-month change in insider ownership percentage.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    institutional_ownership: Optional[float] = Field(
        default=None,
        description="Institutional ownership as a percentage.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    institutional_ownership_change: Optional[float] = Field(
        default=None,
        description="3-month change in institutional ownership percentage.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    price_to_earnings: Optional[float] = Field(
        default=None,
        description="Price to earnings ratio.",
    )
    forward_pe: Optional[float] = Field(
        default=None,
        description="Forward price to earnings ratio.",
    )
    peg_ratio: Optional[float] = Field(
        default=None,
        description="Price/Earnings-To-Growth (PEG) ratio.",
    )
    price_to_sales: Optional[float] = Field(
        default=None,
        description="Price to sales ratio.",
    )
    price_to_book: Optional[float] = Field(
        default=None,
        description="Price to book ratio.",
    )
    price_to_cash: Optional[float] = Field(
        default=None,
        description="Price to cash ratio.",
    )
    price_to_fcf: Optional[float] = Field(
        default=None,
        description="Price to free cash flow ratio.",
    )
    eps_growth_past_1y: Optional[float] = Field(
        default=None,
        description="EPS growth for this year.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    eps_growth_next_1y: Optional[float] = Field(
        default=None,
        description="EPS growth next year.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    eps_growth_past_5y: Optional[float] = Field(
        default=None,
        description="EPS growth for the previous 5 years.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    eps_growth_next_5y: Optional[float] = Field(
        default=None,
        description="EPS growth for the next 5 years.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    sales_growth_past_5y: Optional[float] = Field(
        default=None,
        description="Sales growth for the previous 5 years.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    dividend_yield: Optional[float] = Field(
        default=None,
        description="Annualized dividend yield.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_assets: Optional[float] = Field(
        default=None,
        description="Return on assets.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_equity: Optional[float] = Field(
        default=None,
        description="Return on equity.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_investment: Optional[float] = Field(
        default=None,
        description="Return on investment.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    current_ratio: Optional[float] = Field(
        default=None,
        description="Current ratio.",
    )
    quick_ratio: Optional[float] = Field(
        default=None,
        description="Quick ratio.",
    )
    long_term_debt_to_equity: Optional[float] = Field(
        default=None,
        description="Long term debt to equity ratio.",
    )
    debt_to_equity: Optional[float] = Field(
        default=None,
        description="Total debt to equity ratio.",
    )
    gross_margin: Optional[float] = Field(
        default=None,
        description="Gross margin.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    operating_margin: Optional[float] = Field(
        default=None,
        description="Operating margin.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    profit_margin: Optional[float] = Field(
        default=None,
        description="Profit margin.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )


class FinvizEquityScreenerFetcher(
    Fetcher[FinvizEquityScreenerQueryParams, List[FinvizEquityScreenerData]]
):
    """Finviz Equity Screener Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinvizEquityScreenerQueryParams:
        """Transform query parameters."""
        transformed_params = params.copy()
        if not transformed_params.get("filters_dict"):
            transformed_params["filters_dict"] = {}
        return FinvizEquityScreenerQueryParams(**transformed_params)

    @staticmethod
    def extract_data(  # noqa = PRL0912
        query: FinvizEquityScreenerQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract data from Finviz."""
        # pylint: disable=import-outside-toplevel
        import configparser  # noqa
        from finvizfinance import util
        from finvizfinance.screener import (
            financial,
            overview,
            ownership,
            performance,
            technical,
            valuation,
        )
        from numpy import nan
        from openbb_core.provider.utils.helpers import get_requests_session
        from openbb_finviz.utils.screener_helper import (
            get_preset_choices,
            d_check_screener,
            d_signals,
        )
        from pandas import DataFrame

        preset = None
        util.session = get_requests_session()

        try:
            data_dir = kwargs.get("preferences", {}).get("data_directory")
            preset_choices = get_preset_choices(data_dir)
            preset = query.preset
            if preset is not None and preset not in preset_choices:
                raise OpenBBError(
                    f"Invalid preset '{preset}'. Available presets are:\n{list(preset_choices)}"
                )
        except Exception as e:
            if preset is not None:
                raise e from e
            warn(f"Error loading presets -> {e.__class__.__name__}: {e}")
            preset = None

        data_type = query.metric
        ascend = False
        limit = query.limit
        sleep = 0.1  # For optimized pagination speed without creating too many requests error from Finviz.
        sort_by = "Change"
        df_screen = DataFrame()
        screen_type = {
            "overview": overview.Overview,
            "valuation": valuation.Valuation,
            "financial": financial.Financial,
            "ownership": ownership.Ownership,
            "performance": performance.Performance,
            "technical": technical.Technical,
        }

        if data_type in screen_type:
            screen = screen_type[data_type]()

        if preset is not None:
            preset_filter = configparser.RawConfigParser()
            preset_filter.optionxform = str  # type: ignore
            preset_filter.read(preset_choices[preset])
            d_general = preset_filter["General"]
            d_filters = {
                **preset_filter["Descriptive"],
                **preset_filter["Fundamental"],
                **preset_filter["Technical"],
            }
            for section in ["General", "Descriptive", "Fundamental", "Technical"]:
                for key, val in {**preset_filter[section]}.items():
                    if key not in d_check_screener:
                        raise OpenBBError(
                            f"The screener variable {section}.{key} shouldn't exist!\n"
                        )

                    if val not in d_check_screener[key]:
                        raise OpenBBError(
                            f"Invalid [{section}] {key}={val}. "
                            f"Choose one of the following options:\n{', '.join(d_check_screener[key])}.\n"
                        )

            d_filters = {k: v for k, v in d_filters.items() if v is not None}
            screen.set_filter(filters_dict=d_filters)
            asc = None

            asc = d_general.get("Ascend")

            if asc is not None:
                ascend = asc == "true"

            df_screen = screen.screener_view(
                order=d_general.get("Order", "Change"),
                limit=limit if limit else 100000,
                ascend=ascend,
                sleep_sec=sleep,
                verbose=0,
            )
        # If no preset is supplied, then set the filters based on the query parameters.
        else:
            if query.signal is not None:
                screen.set_filter(signal=d_signals[query.signal])
                if query.signal in ["unusual_volume", "most_active"]:
                    sort_by = "Relative Volume"
                elif query.signal == "top_losers":
                    ascend = True
                elif query.signal in ["new_low", "multiple_bottom", "double_bottom"]:
                    sort_by = "52-Week Low (Relative)"
                    ascend = True
                elif query.signal in ["new_high", "multiple_top", "double_top"]:
                    sort_by = "52-Week High (Relative)"
                elif query.signal == "oversold":
                    sort_by = "Relative Strength Index (14)"
                    ascend = True
                elif query.signal == "overbought":
                    sort_by = "Relative Strength Index (14)"
                elif query.signal == "most_volatile":
                    sort_by = "Volatility (Week)"
                else:
                    pass

            filters_dict: Dict = {}

            if query.sector != "all":
                filters_dict["Sector"] = SECTOR_MAP[query.sector]

            if query.industry != "all":
                filters_dict["Industry"] = INDUSTRY_MAP[query.industry]  # type: ignore

            if query.exchange != "all":
                filters_dict["Exchange"] = EXCHANGE_MAP[query.exchange]

            if query.index != "all":
                filters_dict["Index"] = INDEX_MAP[query.index]

            if query.recommendation != "all":
                filters_dict["Analyst Recom."] = RECOMMENDATION_MAP[
                    query.recommendation
                ]

            if query.mktcap != "all":
                filters_dict["Market Cap."] = MARKET_CAP_MAP[query.mktcap]

            if query.filters_dict is not None:
                _filters_dict = query.filters_dict.copy()  # type: ignore
                order = _filters_dict.pop("Order", None)
                asc = _filters_dict.pop("Ascend", None)
                if asc:
                    ascend = asc == "true"
                if order:
                    sort_by = order
                filters_dict.update(_filters_dict)

            if filters_dict:
                screen.set_filter(filters_dict=filters_dict)

            if not filters_dict and query.signal is None:
                screen.set_filter(signal=d_signals["top_gainers"])
                warn(
                    "No filters or signal provided. Defaulting to 'top_gainers' signal."
                    + " Use the preset, 'all_stocks', to explicitly return every stock on Finviz."
                    + " Returning 10K symbols can take several minutes."
                )

            df_screen = screen.screener_view(
                order=sort_by,
                limit=limit if limit else 100000,
                ascend=ascend,
                sleep_sec=sleep,
                verbose=0,
            )

        if df_screen is None or df_screen.empty:
            raise EmptyDataError(
                "No tickers found for the supplied parameters. Try relaxing the constraints."
            )

        df_screen.columns = [val.strip("\n") for val in df_screen.columns]
        # Commas in the company name can cause issues with delimiters.
        if "Company" in df_screen.columns:
            df_screen["Company"] = df_screen["Company"].str.replace(",", "")

        return df_screen.convert_dtypes().replace({nan: None}).to_dict(orient="records")

    @staticmethod
    def transform_data(
        query: FinvizEquityScreenerQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[FinvizEquityScreenerData]:
        """Transform data."""
        return [FinvizEquityScreenerData.model_validate(d) for d in data]



"""Finviz Price Target Model."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Optional

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.price_target import (
    PriceTargetData,
    PriceTargetQueryParams,
)
from pydantic import Field


class FinvizPriceTargetQueryParams(PriceTargetQueryParams):
    """Finviz Price Target Query.

    Source: https://finviz.com/quote.ashx?
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FinvizPriceTargetData(PriceTargetData):
    """Finviz Price Target Data."""

    __alias_dict__ = {
        "published_date": "Date",
        "status": "Status",
        "analyst_company": "Outer",
        "rating_change": "Rating",
    }

    status: Optional[str] = Field(
        default=None,
        description="The action taken by the firm."
        + " This could be 'Upgrade', 'Downgrade', 'Reiterated', etc.",
    )
    rating_change: Optional[str] = Field(
        default=None,
        description="The rating given by the analyst."
        + " This could be 'Buy', 'Sell', 'Underweight', etc."
        + " If the rating is a revision, the change is indicated by '->'",
    )


class FinvizPriceTargetFetcher(
    Fetcher[FinvizPriceTargetQueryParams, List[FinvizPriceTargetData]]
):
    """Finviz Price Target Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinvizPriceTargetQueryParams:
        """Transform the query params."""
        return FinvizPriceTargetQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FinvizPriceTargetQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the Finviz endpoint."""
        # pylint: disable=import-outside-toplevel
        from finvizfinance import util  # noqa
        from finvizfinance.quote import finvizfinance
        from openbb_core.app.model.abstract.error import OpenBBError
        from openbb_core.provider.utils.errors import EmptyDataError
        from openbb_core.provider.utils.helpers import get_requests_session
        from pandas import DataFrame
        from warnings import warn

        results: List[Dict] = []
        util.session = get_requests_session()
        messages: List = []

        def get_one(symbol) -> List[Dict]:
            """Get the data for one symbol."""
            price_targets = DataFrame()
            result: List[Dict] = []
            try:
                data = finvizfinance(symbol)
                price_targets = data.ticker_outer_ratings()
                if price_targets is None or len(price_targets) == 0:
                    messages.append(f"Failed to get data for {symbol}")
                    return result
                price_targets["symbol"] = symbol
                prices = (
                    price_targets["Price"].astype(str).str.replace("$", "", regex=False)
                )
                price_targets["price_target"] = (
                    prices.str.split("→").str.get(0).str.strip()
                )
                price_targets["adj_price_target"] = (
                    prices.str.split("→").str.get(-1).str.strip()
                )
                price_targets.loc[
                    price_targets["adj_price_target"] == price_targets["price_target"],
                    "price_target",
                ] = None
                price_targets = price_targets.replace("", None).drop(columns="Price")
            except Exception as e:  # pylint: disable=W0718
                messages.append(f"Failed to get data for {symbol} -> {e}")
                return result
            result = price_targets.to_dict(orient="records")
            return result

        symbols = query.symbol.split(",") if query.symbol else []

        for symbol in symbols:
            result = get_one(symbol)
            if result:
                results.extend(result)

        if not results and messages:
            raise OpenBBError("\n".join(messages))

        if not results and not messages:
            raise EmptyDataError("No data was returned for any symbol")

        if results and messages:
            for message in messages:
                warn(message)

        return results

    @staticmethod
    def transform_data(
        query: FinvizPriceTargetQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[FinvizPriceTargetData]:
        """Transform and validate the raw data."""
        return [FinvizPriceTargetData.model_validate(d) for d in data]



"""Finviz Definitions."""

from typing import Literal

GROUPS = Literal[
    "sector",
    "industry",
    "country",
    "capitalization",
    "energy",
    "materials",
    "industrials",
    "consumer_cyclical",
    "consumer_defensive",
    "healthcare",
    "financial",
    "technology",
    "communication_services",
    "utilities",
    "real_estate",
]

GROUPS_DICT = {
    "sector": "Sector",
    "industry": "Industry",
    "country": "Country (U.S. listed stocks only)",
    "capitalization": "Capitalization",
    "energy": "Industry (Energy)",
    "materials": "Industry (Basic Materials)",
    "industrials": "Industry (Industrials)",
    "consumer_cyclical": "Industry (Consumer Cyclical)",
    "consumer_defensive": "Industry (Consumer Defensive)",
    "healthcare": "Industry (Healthcare)",
    "financial": "Industry (Financial)",
    "technology": "Industry (Technology)",
    "communication_services": "Industry (Communication Services)",
    "utilities": "Industry (Utilities)",
    "real_estate": "Industry (Real Estate)",
}

METRICS = Literal[
    "performance",
    "valuation",
    "overview",
]



"""Finviz Compare Groups Model."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Optional

from openbb_core.provider.abstract.data import ForceInt
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.compare_groups import (
    CompareGroupsData,
    CompareGroupsQueryParams,
)
from openbb_core.provider.utils.descriptions import DATA_DESCRIPTIONS
from openbb_finviz.utils.definitions import GROUPS, GROUPS_DICT, METRICS
from pydantic import Field, field_validator

GROUPS_CHOICES = sorted(list(GROUPS_DICT))


class FinvizCompareGroupsQueryParams(CompareGroupsQueryParams):
    """Finviz Compare Groups Query Params."""

    __json_schema_extra__ = {
        "group": {
            "multiple_items_allowed": False,
            "choices": GROUPS_CHOICES,
        },
        "metric": {
            "multiple_items_allowed": False,
            "choices": ["performance", "valuation", "overview"],
        },
    }

    group: GROUPS = Field(
        default="sector",
        description="US-listed stocks only."
        + " When an individual sector is selected, it is broken down by industry."
        + " The default is 'sector'.",
    )
    metric: METRICS = Field(
        default="performance",
        description="Statistical metric to return. Select from: ['performance', 'valuation', 'overview']"
        + " The default is 'performance'.",
    )


class FinvizCompareGroupsData(CompareGroupsData):
    """Finviz Compare Groups Data."""

    __alias_dict__ = {
        "name": "Name",
        "stocks": "Number of Stocks",
        "market_cap": "Market Cap",
        "performance_1d": "Change",
        "performance_1w": "Perf Week",
        "performance_1m": "Perf Month",
        "performance_3m": "Perf Quart",
        "performance_6m": "Perf Half",
        "performance_1y": "Perf Year",
        "performance_ytd": "Perf YTD",
        "volume": "Volume",
        "volume_average": "Avg Volume",
        "volume_relative": "Rel Volume",
        "pe": "P/E",
        "forward_pe": "Fwd P/E",
        "peg": "PEG",
        "eps_growth_past_5y": "EPS past 5Y",
        "eps_growth_next_5y": "EPS next 5Y",
        "sales_growth_past_5_years": "Sales past 5Y",
        "price_to_sales": "P/S",
        "price_to_book": "P/B",
        "price_to_cash": "P/C",
        "price_to_free_cash_flow": "P/FCF",
        "dividend_yield": "Dividend",
        "float_short": "Float Short",
        "analyst_recommendation": "Recom",
    }

    name: str = Field(description="Name or label of the group.")

    stocks: Optional[int] = Field(
        default=None,
        description="The number of stocks in the group.",
    )
    market_cap: Optional[ForceInt] = Field(
        default=None,
        description="The market cap of the group.",
    )
    performance_1d: Optional[float] = Field(
        default=None,
        description="The performance in the last day, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    performance_1w: Optional[float] = Field(
        default=None,
        description="The performance in the last week, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    performance_1m: Optional[float] = Field(
        default=None,
        description="The performance in the last month, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    performance_3m: Optional[float] = Field(
        default=None,
        description="The performance in the last quarter, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    performance_6m: Optional[float] = Field(
        default=None,
        description="The performance in the last half year, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    performance_1y: Optional[float] = Field(
        default=None,
        description="The performance in the last year, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    performance_ytd: Optional[float] = Field(
        default=None,
        description="The performance in the year to date, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    dividend_yield: Optional[float] = Field(
        default=None,
        description="The dividend yield of the group, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    pe: Optional[float] = Field(
        default=None,
        description="The P/E ratio of the group.",
    )
    forward_pe: Optional[float] = Field(
        default=None,
        description="The forward P/E ratio of the group.",
    )
    peg: Optional[float] = Field(
        default=None,
        description="The PEG ratio of the group.",
    )
    eps_growth_past_5y: Optional[float] = Field(
        default=None,
        description="The EPS growth of the group for the past 5 years, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    eps_growth_next_5y: Optional[float] = Field(
        default=None,
        description="The estimated EPS growth of the groupo for the next 5 years,"
        + " as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    sales_growth_past_5y: Optional[float] = Field(
        default=None,
        description="The sales growth of the group for the past 5 years, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    float_short: Optional[float] = Field(
        default=None,
        description="The percent of the float shorted for the group, as a normalized value.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    analyst_recommendation: Optional[float] = Field(
        default=None,
        description="The analyst consensus, on a scale of 1-5 where 1 is a buy and 5 is a sell.",
    )
    volume: Optional[ForceInt] = Field(
        default=None, description=DATA_DESCRIPTIONS.get("volume", "")
    )
    volume_average: Optional[ForceInt] = Field(
        default=None,
        description="The 3-month average volume of the group.",
    )
    volume_relative: Optional[float] = Field(
        default=None,
        description="The relative volume compared to the 3-month average volume.",
    )

    @field_validator(
        "performance_1w",
        "performance_1m",
        "performance_3m",
        "performance_6m",
        "performance_1y",
        "performance_ytd",
        "dividend_yield",
        "eps_growth_past_5y",
        "eps_growth_next_5y",
        "sales_growth_past_5y",
        "float_short",
        mode="before",
        check_fields=False,
    )
    @classmethod
    def validate_percent_values(cls, v):
        """Check fields for the presence of a % character."""
        if v is not None and "%" in str(v):
            return float(v.replace("%", "")) / 100
        return v if v else None

    @field_validator("market_cap", "volume", mode="before", check_fields=False)
    @classmethod
    def validate_abbreviated_numbers(cls, v):
        """Check for abbreviated string values."""
        if v is not None and isinstance(v, str):
            v = (
                v.replace("M", "e+6")
                .replace("B", "e+9")
                .replace("T", "e+12")
                .replace("K", "e+3")
            )
            return int(float(v))
        return v if v else None


class FinvizCompareGroupsFetcher(
    Fetcher[FinvizCompareGroupsQueryParams, List[FinvizCompareGroupsData]]
):
    """Finviz Compare Groups Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinvizCompareGroupsQueryParams:
        """Transform the query params."""
        if params.get("group") is None:
            params["group"] = "sector"
        if params.get("metric") is None:
            params["metric"] = "performance"
        return FinvizCompareGroupsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FinvizCompareGroupsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the raw data from Finviz."""
        # pylint: disable=import-outside-toplevel
        from finvizfinance import util
        from finvizfinance.group import Overview, Performance, Valuation
        from openbb_core.provider.utils.helpers import get_requests_session
        from pandas import DataFrame

        util.session = get_requests_session()
        results: List = []
        data = DataFrame()
        if query.metric == "performance":
            data = Performance().screener_view(
                group=GROUPS_DICT[query.group],  # type: ignore
                order="Performance (Week)",
            )
        if query.metric == "valuation":
            data = Valuation().screener_view(
                group=GROUPS_DICT[query.group],  # type: ignore
                order="Forward Price/Earnings",
            )
        if query.metric == "overview":
            data = Overview().screener_view(
                group=GROUPS_DICT[query.group],  # type: ignore
                order="Change",
            )

        if not data.empty:
            results = data.fillna("N/A").replace("N/A", None).to_dict(orient="records")

        return results

    @staticmethod
    def transform_data(
        query: FinvizCompareGroupsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[FinvizCompareGroupsData]:
        """Transform the raw data."""
        return [FinvizCompareGroupsData.model_validate(d) for d in data]



"""Finviz Equity Profile Model."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Optional
from warnings import warn

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.equity_info import (
    EquityInfoData,
    EquityInfoQueryParams,
)
from pydantic import Field


class FinvizEquityProfileQueryParams(EquityInfoQueryParams):
    """
    Finviz Equity Profile Query.

    Source: https://finviz.com/screener.ashx
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FinvizEquityProfileData(EquityInfoData):
    """Finviz Equity Profile Data."""

    __alias_dict__ = {
        "stock_exchange": "exchange",
    }

    index: Optional[str] = Field(
        default=None,
        description="Included in indices - i.e., Dow, Nasdaq, or S&P.",
    )
    optionable: Optional[str] = Field(
        default=None,
        description="Whether options trade against the ticker.",
    )
    shortable: Optional[str] = Field(
        default=None,
        description="If the asset is shortable.",
    )
    shares_outstanding: Optional[str] = Field(
        default=None,
        description="The number of shares outstanding, as an abbreviated string.",
    )
    shares_float: Optional[str] = Field(
        default=None,
        description="The number of shares in the public float, as an abbreviated string.",
    )
    short_interest: Optional[str] = Field(
        default=None,
        description="The last reported number of shares sold short, as an abbreviated string.",
    )
    institutional_ownership: Optional[float] = Field(
        default=None,
        description="The institutional ownership of the stock, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    market_cap: Optional[str] = Field(
        default=None,
        description="The market capitalization of the stock, as an abbreviated string.",
    )
    dividend_yield: Optional[float] = Field(
        default=None,
        description="The dividend yield of the stock, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    earnings_date: Optional[str] = Field(
        default=None,
        description="The last, or next confirmed, earnings date and announcement time, as a string."
        + " The format is Nov 02 AMC - for after market close.",
    )
    beta: Optional[float] = Field(
        default=None,
        description="The beta of the stock relative to the broad market.",
    )


class FinvizEquityProfileFetcher(
    Fetcher[FinvizEquityProfileQueryParams, List[FinvizEquityProfileData]]
):
    """Finviz Equity Profile Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinvizEquityProfileQueryParams:
        """Transform the query params."""
        return FinvizEquityProfileQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FinvizEquityProfileQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the raw data from Finviz."""
        # pylint: disable=import-outside-toplevel
        from finvizfinance import util
        from finvizfinance.quote import finvizfinance
        from openbb_core.app.model.abstract.error import OpenBBError
        from openbb_core.provider.utils.errors import EmptyDataError
        from openbb_core.provider.utils.helpers import get_requests_session

        util.session = get_requests_session()
        results: List = []
        messages: List = []

        def get_one(symbol) -> Dict:
            """Get the data for one symbol."""
            result: Dict = {}
            try:
                data = finvizfinance(symbol)
                fundament = data.ticker_fundament()
                description = data.ticker_description()
            except Exception as e:  # pylint: disable=W0718
                messages.append(f"Failed to get data for {symbol} -> {e}")
                return result
            div_yield = (
                float(str(fundament.get("Dividend %", None)).replace("%", "")) / 100
                if fundament.get("Dividend %", "-") != "-"
                else None
            )
            inst_own = (
                float(str(fundament.get("Inst Own", None)).replace("%", "")) / 100
                if fundament.get("Inst Own", "-") != "-"
                else None
            )
            result.update(
                {
                    "symbol": symbol,
                    "exchange": (
                        fundament.get("Exchange", None)
                        if fundament.get("Exchange", "-") != "-"
                        else None
                    ),
                    "name": (
                        fundament.get("Company", None)
                        if fundament.get("Company", "-") != "-"
                        else None
                    ),
                    "sector": (
                        fundament.get("Sector", None)
                        if fundament.get("Sector", "-") != "-"
                        else None
                    ),
                    "industry_category": (
                        fundament.get("Industry", None)
                        if fundament.get("Industry", "-") != "-"
                        else None
                    ),
                    "hq_country": (
                        fundament.get("Country", None)
                        if fundament.get("Country", "-") != "-"
                        else None
                    ),
                    "employees": (
                        fundament.get("Employees", None)
                        if fundament.get("Employees", "-") != "-"
                        else None
                    ),
                    "index": (
                        fundament.get("Index", None)
                        if fundament.get("Index", "-") != "-"
                        else None
                    ),
                    "beta": (
                        fundament.get("Beta", None)
                        if fundament.get("Beta", "-") != "-"
                        else None
                    ),
                    "optionable": (
                        fundament.get("Optionable", None)
                        if fundament.get("Optionable", "-") != "-"
                        else None
                    ),
                    "shortable": (
                        fundament.get("Shortable", None)
                        if fundament.get("Shortable", "-") != "-"
                        else None
                    ),
                    "shares_outstanding": (
                        fundament.get("Shs Outstand", None)
                        if fundament.get("Shs Outstand", "-") != "-"
                        else None
                    ),
                    "shares_float": (
                        fundament.get("Shs Float", None)
                        if fundament.get("Shs Float", "-") != "-"
                        else None
                    ),
                    "short_interest": (
                        fundament.get("Short Interest", None)
                        if fundament.get("Short Interest", "-") != "-"
                        else None
                    ),
                    "institutional_ownership": inst_own if inst_own else None,
                    "market_cap": (
                        fundament.get("Market Cap", None)
                        if fundament.get("Market Cap", "-") != "-"
                        else None
                    ),
                    "dividend_yield": div_yield if div_yield else None,
                    "earnings_date": (
                        fundament.get("Earnings", None)
                        if fundament.get("Earnings", "-") != "-"
                        else None
                    ),
                    "long_description": description if description else None,
                }
            )

            return result

        symbols = query.symbol.split(",")
        for symbol in symbols:
            result = get_one(symbol)
            if result is not None and result:
                results.append(result)

        if not results and messages:
            raise OpenBBError("\n".join(messages))

        if not results and not messages:
            raise EmptyDataError("No data was returned for any symbol")

        if results and messages:
            for message in messages:
                warn(message)

        return results

    @staticmethod
    def transform_data(
        query: FinvizEquityProfileQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[FinvizEquityProfileData]:
        """Transform and validate the raw data."""
        return [FinvizEquityProfileData.model_validate(d) for d in data]



"""Screener Helpers """

# pylint: disable=too-many-lines

from typing import Dict, Literal

MARKET_CAP_MAP = {
    "all": "Any",
    "mega": "Mega ($200bln and more)",
    "large": "Large ($10bln to $200bln)",
    "large_over": "+Large (over $10bln)",
    "large_under": "-Large (under $200bln)",
    "mid": "Mid ($2bl to $10bln)",
    "mid_over": "+Mid (over $2bln)",
    "mid_under": "-Mid (under $10bln)",
    "small": "Small ($300mln to $2bln)",
    "small_over": "+Small (over $300mln)",
    "small_under": "-Small (under $2bln)",
    "micro": "Micro ($50mln to $300mln)",
    "micro_over": "+Micro (over $50mln)",
    "micro_under": "-Micro (under $300mln)",
    "nano": "Nano (under $50mln)",
}

MarketCap = Literal[
    "all",
    "mega",
    "large",
    "large_over",
    "large_under",
    "mid",
    "mid_over",
    "mid_under",
    "small",
    "small_over",
    "small_under",
    "micro",
    "micro_over",
    "micro_under",
    "nano",
]

INDUSTRY_MAP = {
    "all": "Any",
    "stocks_only": "Stocks only (ex-Funds)",
    "etf": "Exchange Traded Fund",
    "advertising_agencies": "Advertising Agencies",
    "aerospace_defense": "Aerospace & Defense",
    "agricultural_inputs": "Agricultural Inputs",
    "airlines": "Airlines",
    "airports_airservices": "Airports & Air Services",
    "aluminum": "Aluminum",
    "apparel_manufacturing": "Apparel Manufacturing",
    "apparel_retail": "Apparel Retail",
    "asset_management": "Asset Management",
    "auto_manufacturers": "Auto Manufacturers",
    "auto_parts": "Auto Parts",
    "auto_dealerships": "Auto & Truck Dealerships",
    "banks_diversified": "Banks - Diversified",
    "banks_regional": "Banks - Regional",
    "beverages_brewers": "Beverages - Brewers",
    "beverages_nonalcoholic": "Beverages - Non-Alcoholic",
    "beverages_wineries_distilleries": "Beverages - Wineries & Distilleries",
    "biotechnology": "Biotechnology",
    "broadcasting": "Broadcasting",
    "building_materials": "Building Materials",
    "building_products_equipment": "Building Products & Equipment",
    "business_equipment_supplies": "Business Equipment & Supplies",
    "capital_markets": "Capital Markets",
    "chemicals": "Chemicals",
    "closed_end_fund_debt": "Closed-End Fund - Debt",
    "closed_end_fund_equity": "Closed-End Fund - Equity",
    "closed_end_fund_foreign": "Closed-End Fund - Foreign",
    "coking_coal": "Coking Coal",
    "communication_equipment": "Communication Equipment",
    "computer_hardware": "Computer Hardware",
    "confectioners": "Confectioners",
    "conglomerates": "Conglomerates",
    "consulting_services": "Consulting Services",
    "consumer_electronics": "Consumer Electronics",
    "copper": "Copper",
    "credit_services": "Credit Services",
    "department_stores": "Department Stores",
    "diagnostics_research": "Diagnostics & Research",
    "discount_stores": "Discount Stores",
    "drug_manufacturers_general": "Drug Manufacturers - General",
    "drug_manufacturers_specialty_generic": "Drug Manufacturers - Specialty & Generic",
    "education_training_services": "Education & Training Services",
    "electrical_equipment_parts": "Electrical Equipment & Parts",
    "electronic_components": "Electronic Components",
    "electronic_gaming_multimedia": "Electronic Gaming & Multimedia",
    "electronics_computer_distribution": "Electronics & Computer Distribution",
    "engineering_construction": "Engineering & Construction",
    "entertainment": "Entertainment",
    "farm_heavy_construction_machinery": "Farm & Heavy Construction Machinery",
    "farm_products": "Farm Products",
    "financial_conglomerates": "Financial Conglomerates",
    "financial_data_stock_exchanges": "Financial Data & Stock Exchanges",
    "food_distribution": "Food Distribution",
    "footwear_accessories": "Footwear & Accessories",
    "furnishings_fixtures_appliances": "Furnishings, Fixtures & Appliances",
    "gambling": "Gambling",
    "gold": "Gold",
    "grocery_stores": "Grocery Stores",
    "health_care_plans": "Healthcare Plans",
    "health_information_services": "Health Information Services",
    "home_improvement_retail": "Home Improvement Retail",
    "household_personal_products": "Household & Personal Products",
    "industrial_distribution": "Industrial Distribution",
    "information_technology_services": "Information Technology Services",
    "infrastructure_operations": "Infrastructure Operations",
    "insurance_brokers": "Insurance Brokers",
    "insurance_diversified": "Insurance - Diversified",
    "insurance_life": "Insurance - Life",
    "insurance_property_casualty": "Insurance - Property & Casualty",
    "insurance_reinsurance": "Insurance - Reinsurance",
    "insurance_specialty": "Insurance - Specialty",
    "integrated_freight_logistics": "Integrated Freight & Logistics",
    "internet_content_information": "Internet Content & Information",
    "internet_retail": "Internet Retail",
    "leisure": "Leisure",
    "lodging": "Lodging",
    "lumber_wood_production": "Lumber & Wood Production",
    "luxury_goods": "Luxury Goods",
    "marine_shipping": "Marine Shipping",
    "medical_care_facilities": "Medical Care Facilities",
    "medical_devices": "Medical Devices",
    "medical_distribution": "Medical Distribution",
    "medical_instruments_supplies": "Medical Instruments & Supplies",
    "metal_fabrication": "Metal Fabrication",
    "mortgage_finance": "Mortgage Finance",
    "oil_gas_drilling": "Oil & Gas Drilling",
    "oil_gas_ep": "Oil & Gas E&P",
    "oil_gas_equipment_services": "Oil & Gas Equipment & Services",
    "oil_gas_integrated": "Oil & Gas Integrated",
    "oil_gas_midstream": "Oil & Gas Midstream",
    "oil_gas_refining_marketing": "Oil & Gas Refining & Marketing",
    "other_industrial_metals_mining": "Other Industrial Metals & Mining",
    "other_precious_metals_mining": "Other Precious Metals & Mining",
    "packaged_foods": "Packaged Foods",
    "packaging_containers": "Packaging & Containers",
    "paper_paper_products": "Paper & Paper Products",
    "personal_services": "Personal Services",
    "pharmaceutical_retailers": "Pharmaceutical Retailers",
    "pollution_treatment_controls": "Pollution & Treatment Controls",
    "publishing": "Publishing",
    "railroads": "Railroads",
    "real_estate_development": "Real Estate - Development",
    "real_estate_diversified": "Real Estate - Diversified",
    "real_estate_services": "Real Estate Services",
    "recreational_vehicles": "Recreational Vehicles",
    "reit_diversified": "REIT - Diversified",
    "reit_health_care_facilities": "REIT - Healthcare Facilities",
    "reit_hotel_motel": "REIT - Hotel & Motel",
    "reit_industrial": "REIT - Industrial",
    "reit_mortgage": "REIT - Mortgage",
    "reit_office": "REIT - Office",
    "reit_residential": "REIT - Residential",
    "reit_retail": "REIT - Retail",
    "reit_specialty": "REIT - Specialty",
    "rental_leasing_services": "Rental & Leasing Services",
    "residential_construction": "Residential Construction",
    "resorts_casinos": "Resorts & Casinos",
    "restaurants": "Restaurants",
    "scientific_technical_instruments": "Scientific & Technical Instruments",
    "security_protection_services": "Security & Protection Services",
    "semiconductor_equipment_materials": "Semiconductor Equipment & Materials",
    "semiconductors": "Semiconductors",
    "shell_companies": "Shell Companies",
    "silver": "Silver",
    "software_application": "Software - Application",
    "software_infrastructure": "Software - Infrastructure",
    "solar": "Solar",
    "specialty_business_services": "Specialty Business Services",
    "specialty_chemicals": "Specialty Chemicals",
    "specialty_industrial_machinery": "Specialty Industrial Machinery",
    "specialty_retail": "Specialty Retail",
    "staffing_employment_services": "Staffing & Employment Services",
    "steel": "Steel",
    "telecom_services": "Telecom Services",
    "textile_manufacturing": "Textile Manufacturing",
    "thermal_coal": "Thermal Coal",
    "tobacco": "Tobacco",
    "tools_accessories": "Tools & Accessories",
    "travel_services": "Travel Services",
    "trucking": "Trucking",
    "uranium": "Uranium",
    "utilities_diversified": "Utilities - Diversified",
    "utilities_independent_power_producers": "Utilities - Independent Power Producers",
    "utilities_regulated_electric": "Utilities - Regulated Electric",
    "utilities_regulated_gas": "Utilities - Regulated Gas",
    "utilities_regulated_water": "Utilities - Regulated Water",
    "utilities_renewable": "Utilities - Renewable",
    "waste_management": "Waste Management",
}

d_signals = {
    "top_gainers": "Top Gainers",
    "top_losers": "Top Losers",
    "new_high": "New High",
    "new_low": "New Low",
    "most_volatile": "Most Volatile",
    "most_active": "Most Active",
    "unusual_volume": "Unusual Volume",
    "overbought": "Overbought",
    "oversold": "Oversold",
    "downgrades": "Downgrades",
    "upgrades": "Upgrades",
    "earnings_before": "Earnings Before",
    "earnings_after": "Earnings After",
    "recent_insider_buying": "Recent Insider Buying",
    "recent_insider_selling": "Recent Insider Selling",
    "major_news": "Major News",
    "horizontal_sr": "Horizontal S/R",
    "tl_resistance": "TL Resistance",
    "tl_support": "TL Support",
    "wedge_up": "Wedge Up",
    "wedge_down": "Wedge Down",
    "wedge": "Wedge",
    "triangle_ascending": "Triangle Ascending",
    "triangle_descending": "Triangle Descending",
    "channel_up": "Channel Up",
    "channel_down": "Channel Down",
    "channel": "Channel",
    "double_top": "Double Top",
    "double_bottom": "Double Bottom",
    "multiple_top": "Multiple Top",
    "multiple_bottom": "Multiple Bottom",
    "head_shoulders": "Head & Shoulders",
    "head_shoulders_inverse": "Head & Shoulders Inverse",
}


d_signals_desc = {
    "top_gainers": "stocks with the highest price gain percent today",
    "top_losers": "stocks with the highest price percent loss today",
    "new_high": "stocks making 52-week high today",
    "new_low": "stocks making 52-week low today",
    "most_volatile": "stocks with the highest widest high/low trading range today",
    "most_active": "stocks with the highest trading volume today",
    "unusual_volume": "stocks with unusually high volume today - the highest relative volume ratio",
    "overbought": "stock is becoming overvalued and may experience a pullback.",
    "oversold": "oversold stocks may represent a buying opportunity for investors",
    "downgrades": "stocks downgraded by analysts today",
    "upgrades": "stocks upgraded by analysts today",
    "earnings_before": "companies reporting earnings today, before market open",
    "earnings_after": "companies reporting earnings today, after market close",
    "recent_insider_buying": "stocks with recent insider buying activity",
    "recent_insider_selling": "stocks with recent insider selling activity",
    "major_news": "stocks with the highest news coverage today",
    "horizontal_sr": "horizontal channel of price range between support and resistance trendlines",
    "tl_resistance": "once a rising trendline is broken",
    "tl_support": "once a falling trendline is broken",
    "wedge_up": "upward trendline support and upward trendline resistance (reversal)",
    "wedge_down": "downward trendline support and downward trendline resistance (reversal)",
    "wedge": "upward trendline support, downward trendline resistance (contiunation)",
    "triangle_ascending": "upward trendline support and horizontal trendline resistance",
    "triangle_descending": "horizontal trendline support and downward trendline resistance",
    "channel_up": "both support and resistance trendlines slope upward",
    "channel_down": "both support and resistance trendlines slope downward",
    "channel": "both support and resistance trendlines are horizontal",
    "double_top": "stock with 'M' shape that indicates a bearish reversal in trend",
    "double_bottom": "stock with 'W' shape that indicates a bullish reversal in trend",
    "multiple_top": "same as double_top hitting more highs",
    "multiple_bottom": "same as double_bottom hitting more lows",
    "head_shoulders": "chart formation that predicts a bullish-to-bearish trend reversal",
    "head_shoulders_inverse": "chart formation that predicts a bearish-to-bullish trend reversal",
}

Recommendation = Literal[
    "all",
    "strong_buy",
    "buy+",
    "buy",
    "hold+",
    "hold",
    "hold-",
    "sell",
    "sell-",
    "strong_sell",
]

RECOMMENDATION_MAP = {
    "all": "Any",
    "strong_buy": "Strong Buy (1)",
    "buy+": "Buy or better",
    "buy": "Buy",
    "hold+": "Hold or better",
    "hold": "Hold",
    "hold-": "Hold or worse",
    "sell": "Sell",
    "sell-": "Sell or worse",
    "strong_sell": "Strong Sell (5)",
}

d_check_screener = {
    "Order": [
        "Any",
        "Signal",
        "Ticker",
        "Company",
        "Sector",
        "Industry",
        "Country",
        "Market Cap.",
        "Price/Earnings",
        "Forward Price/Earnings",
        "PEG (Price/Earnings/Growth)",
        "Price/Sales",
        "Price/Book",
        "Price/Cash",
        "Price/Free Cash Flow",
        "Dividend Yield",
        "Payout Ratio",
        "EPS(ttm)",
        "EPS growth this year",
        "EPS growth next year",
        "EPS growth past 5 years",
        "EPS growth next 5 years",
        "Sales growth past 5 years",
        "EPS growth qtr over qtr",
        "Sales growth qtr over qtr",
        "Shares Outstanding",
        "Shares Float",
        "Insider Ownership",
        "Insider Transactions",
        "Institutional Ownership",
        "Institutional Transactions",
        "Short Interest Share",
        "Short Interest Ratio",
        "Earnings Date",
        "Return on Assets",
        "Return on Equity",
        "Return on Investment",
        "Current Ratio",
        "Quick Ratio",
        "LT Debt/Equity",
        "Total Debt/Equity",
        "Gross Margin",
        "Operating Margin",
        "Net Profit Margin",
        "Analyst Recommendation",
        "Performance (Week)",
        "Performance (Month)",
        "Performance (Quarter)",
        "Performance (Half Year)",
        "Performance (Year)",
        "Performance (Year To Date)",
        "Beta",
        "Average True Range",
        "Volatility (Week)",
        "Volatility (Month)",
        "20-Day SMA (Relative)",
        "50-Day SMA (Relative)",
        "200-Day SMA (Relative)",
        "50-Day High (Relative)",
        "50-Day Low (Relative)",
        "52-Week High (Relative)",
        "52-Week Low (Relative)",
        "Relative Strength Index (14)",
        "Average Volume (3 Month)",
        "Relative Volume",
        "Change",
        "Change from Open",
        "Gap",
        "Volume",
        "Price",
        "Target Price",
        "IPO Date",
    ],
    "Ascend": ["true", "false"],
    "Signal": [
        "Any",
        "Top Gainers",
        "Top Losers",
        "New High",
        "New Low",
        "Most Volatile",
        "Most Active",
        "Unusual Volume",
        "Overbought",
        "Oversold",
        "Downgrades",
        "Upgrades",
        "Earnings Before",
        "Earnings After",
        "Recent Insider Buying",
        "Recent Insider Selling",
        "Major News",
        "Horizontal S/R",
        "TL Resistance",
        "TL Support",
        "Wedge Up",
        "Wedge Down",
        "Triangle Ascending",
        "Triangle Descending",
        "Wedge",
        "Channel Up",
        "Channel Down",
        "Channel",
        "Double Top",
        "Double Bottom",
        "Multiple Top",
        "Multiple Bottom",
        "Head & Shoulders",
        "Head & Shoulders Inverse",
    ],
    "Exchange": ["Any", "AMEX", "NASDAQ", "NYSE"],
    "Market Cap.": [
        "Any",
        "Mega ($200bln and more)",
        "Large ($10bln to $200bln)",
        "Mid ($2bln to $10bln)",
        "Small ($300mln to $2bln)",
        "Micro ($50mln to $300mln)",
        "Nano (under $50mln)",
        "+Large (over $10bln)",
        "+Mid (over $2bln)",
        "+Small (over $300mln)",
        "+Micro (over $50mln)",
        "-Large (under $200bln)",
        "-Mid (under $10bln)",
        "-Small (under $2bln)",
        "-Micro (under $300mln)",
    ],
    "Earnings Date": [
        "Any",
        "Today",
        "Today Before Market Open",
        "Today Before Market Close",
        "Tomorrow",
        "Tomorrow Before Market Open",
        "Tomorrow Before Market Close",
        "Yesterday",
        "Yesterday Before Market Open",
        "Yesterday Before Market Close",
        "Next 5 Days",
        "Previous 5 Days",
        "This Week",
        "Next Week",
        "Previous Week",
        "This Month",
    ],
    "Target Price": [
        "Any",
        "50% Above Price",
        "40% Above Price",
        "30% Above Price",
        "20% Above Price",
        "10% Above Price",
        "5% Above Price",
        "Above Price",
        "Below Price",
        "5% Below Price",
        "10% Below Price",
        "20% Below Price",
        "30% Below Price",
        "40% Below Price",
        "50% Below Price",
    ],
    "Index": ["Any", "S&P 500", "DJIA", "NASDAQ 100", "Russell 2000"],
    "Dividend Yield": [
        "Any",
        "None (0%)",
        "Positive (>0%)",
        "High (>5%)",
        "Very High (>10%)",
        "Over 1%",
        "Over 2%",
        "Over 3%",
        "Over 4%",
        "Over 5%",
        "Over 6%",
        "Over 7%",
        "Over 8%",
        "Over 9%",
        "Over 10%",
    ],
    "Average Volume": [
        "Any",
        "Under 50K",
        "Under 100K",
        "Under 500K",
        "Under 750K",
        "Under 1M",
        "Over 50K",
        "Over 100K",
        "Over 200K",
        "Over 300K",
        "Over 400K",
        "Over 500K",
        "Over 750K",
        "Over 1M",
        "Over 2M",
        "100K to 500K",
        "100K to 1M",
        "500K to 1M",
        "500K to 10M",
    ],
    "IPO Date": [
        "Any",
        "Today",
        "Yesterday",
        "In the last week",
        "In the last month",
        "In the last quarter",
        "In the last year",
        "In the last 2 years",
        "In the last 3 years",
        "In the last 5 years",
        "More than a year ago",
        "More that 5 years ago",
        "More than 10 years ago",
        "More than 15 years ago",
        "More than 20 years ago",
        "More than 25 years ago",
    ],
    "Sector": [
        "Any",
        "Basic Materials",
        "Communication Services",
        "Consumer Cyclical",
        "Consumer Defensive",
        "Energy",
        "Financial",
        "Healthcare",
        "Industrials",
        "Real Estate",
        "Technology",
        "Utilities",
    ],
    "Float Short": [
        "Any",
        "Low (<5%)",
        "High(>20%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "Relative Volume": [
        "Any",
        "Over 10",
        "Over 5",
        "Over 3",
        "Over 2",
        "Over 1.5",
        "Over 1",
        "Over 0.75",
        "Over 0.5",
        "Over 0.25",
        "Under 2",
        "Under 1",
        "Under 1.5",
        "Under 1",
        "Under 0.75",
        "Under 0.5",
        "Under 0.25",
        "Under 0.1",
    ],
    "Shares Outstanding": [
        "Any",
        "Under 1M",
        "Under 5M",
        "Under 10M",
        "Under 20M",
        "Under 50M",
        "Under 100M",
        "Over 1M",
        "Over 2M",
        "Over 5M",
        "Over 10M",
        "Over 20M",
        "Over 50M",
        "Over 100M",
        "Over 200M",
        "Over 500M",
        "Over 1000M",
    ],
    "Industry": list(INDUSTRY_MAP.values()),
    "Analyst Recom.": [
        "Any",
        "Strong Buy (1)",
        "Buy or better",
        "Buy",
        "Hold or better",
        "Hold",
        "Hold or worse",
        "Sell",
        "Sell or worse",
        "Strong Sell (5)",
    ],
    "Current Volume": [
        "Any",
        "Under 100K",
        "Under 500K",
        "Under 750K",
        "Under 1M",
        "Over 0",
        "Over 50K",
        "Over 100K",
        "Over 200K",
        "Over 300K",
        "Over 400K",
        "Over 500K",
        "Over 750K",
        "Over 1M",
        "Over 2M",
        "Over 5M",
        "Over 10M",
        "Over 20M",
    ],
    "Float": [
        "Any",
        "Under 1M",
        "Under 5M",
        "Under 10M",
        "Under 20M",
        "Under 50M",
        "Under 100M",
        "Over 1M",
        "Over 2M",
        "Over 5M",
        "Over 10M",
        "Over 20M",
        "Over 50M",
        "Over 100M",
        "Over 200M",
        "Over 500M",
        "Over 1000M",
    ],
    "Country": [
        "Any",
        "Asia",
        "Europe",
        "Latin America",
        "BRIC",
        "Argentina",
        "Australia",
        "Bahamas",
        "Belgium",
        "BeNeLux",
        "Bermuda",
        "Brazil",
        "Canada",
        "Cayman Islands",
        "Chile",
        "China",
        "China & Hong Kong",
        "Colombia",
        "Cyprus",
        "Denmark",
        "Finland",
        "France",
        "Germany",
        "Greece",
        "Hong Kong",
        "Hungary",
        "Iceland",
        "Iceland",
        "India",
        "Indonesia",
        "Ireland",
        "Israel",
        "Italy",
        "Japan",
        "Kazakhstan",
        "Luxembourg",
        "Malaysia",
        "Malta",
        "Mexico",
        "Monaco",
        "Netherlands",
        "New Zealand",
        "Norway",
        "Panama",
        "Peru",
        "Philippines",
        "Portugal",
        "Russia",
        "Singapore",
        "South Africa",
        "South Korea",
        "Spain",
        "Sweden",
        "Switzerland",
        "Taiwan",
        "Turkey",
        "United Arab Emirates",
        "United Kingdom",
        "Uruguay",
        "USA",
        "Foreign (ex-USA)",
    ],
    "Option/Short": ["Any", "Optionable", "Shortable", "Optionable and shortable"],
    "Price": [
        "Any",
        "Under $1",
        "Under $2",
        "Under $3",
        "Under $4",
        "Under $5",
        "Under $7",
        "Under $10",
        "Under $15",
        "Under $20",
        "Under $30",
        "Under $40",
        "Under $50",
        "Over $1",
        "Over $2",
        "Over $3",
        "Over $4",
        "Over $5",
        "Over $7",
        "Over $10",
        "Over $15",
        "Over $20",
        "Over $30",
        "Over $40",
        "Over $50",
        "Over $60",
        "Over $70",
        "Over $80",
        "Over $90",
        "Over $100",
        "$1 to $5",
        "$1 to $10",
        "$1 to $20",
        "$5 to %10",
        "$5 to $20",
        "$5 to $50",
        "$10 to $20",
        "$10 to $50",
        "$20 to $50",
        "$50 to $100",
    ],
    "P/E": [
        "Any",
        "Low (<15)",
        "Profitable (>0)",
        "High (>50)",
        "Under 5",
        "Under 10",
        "Under 15",
        "Under 20",
        "Under 25",
        "Under 30",
        "Under 35",
        "Under 40",
        "Under 45",
        "Under 50",
        "Over 5",
        "Over 10",
        "Over 15",
        "Over 25",
        "Over 30",
        "Over 35",
        "Over 40",
        "Over 45",
        "Over 50",
    ],
    "Price/Cash": [
        "Any",
        "Low (<3)",
        "High (>50)",
        "Under 1",
        "Under 2",
        "Under 3",
        "Under 4",
        "Under 5",
        "Under 6",
        "Under 7",
        "Under 8",
        "Under 9",
        "Under 10",
        "Over 1",
        "Over 2",
        "Over 3",
        "Over 4",
        "Over 5",
        "Over 6",
        "Over 7",
        "Over 8",
        "Over 9",
        "Over 10",
        "Over 20",
        "Over 30",
        "Over 40",
        "Over 50",
    ],
    "EPS growthnext 5 years": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Positive Low (<10%)",
        "High (>25%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "Return on Equity": [
        "Any",
        "Positive (>0%)",
        "Negative (<0%)",
        "Very Positive (>30%)",
        "Very Negative (<-15%)",
        "Under -50%",
        "Under -40%",
        "Under -35%",
        "Under -30%",
        "Under -25%",
        "Under -20%",
        "Under -15%",
        "Under -10%",
        "Under -5%",
        "Over +50%",
        "Over +45%",
        "Over +40%",
        "Over +35%",
        "Over +30%",
        "Over +25%",
        "Over +20%",
        "Over +15%",
        "Over +10%",
        "Over +5%",
    ],
    "Debt/Equity": [
        "Any",
        "High (>0.5)",
        "Low (<0.1)",
        "Under 1",
        "Under 0.9",
        "Under 0.8",
        "Under 0.7",
        "Under 0.6",
        "Under 0.5",
        "Under 0.4",
        "Under 0.3",
        "Under 0.2",
        "Under 0.1",
        "Over 0.1",
        "Over 0.2",
        "Over 0.3",
        "Over 0.4",
        "Over 0.5",
        "Over 0.6",
        "Over 0.7",
        "Over 0.8",
        "Over 0.9",
        "Over 1",
    ],
    "InsiderOwnership": [
        "Any",
        "Low (<5%)",
        "High (>30%)",
        "Very High (>50%)",
        "Over 10%",
        "Over 20%",
        "Over 30%",
        "Over 40%",
        "Over 50%",
        "Over 60%",
        "Over 70%",
        "Over 80%",
        "Over 90%",
    ],
    "Forward P/E": [
        "Any",
        "Low (<15)",
        "Profitable (>0)",
        "High (>50)",
        "Under 5",
        "Under 10",
        "Under 15",
        "Under 20",
        "Under 25",
        "Under 30",
        "Under 35",
        "Under 40",
        "Under 45",
        "Under 50",
        "Over 5",
        "Over 10",
        "Over 15",
        "Over 20",
        "Over 25",
        "Over 30",
        "Over 35",
        "Over 40",
        "Over 45",
        "Over 50",
    ],
    "Price/Free Cash Flow": [
        "Any",
        "Low (<15)",
        "High (>50)",
        "Under 5",
        "Under 10",
        "Under 15",
        "Under 20",
        "Under 25",
        "Under 30",
        "Under 35",
        "Under 40",
        "Under 45",
        "Under 50",
        "Under 60",
        "Under 70",
        "Under 80",
        "Under 90",
        "Under 100",
        "Over 5",
        "Over 10",
        "Over 15",
        "Over 20",
        "Over 25",
        "Over 30",
        "Over 35",
        "Over 40",
        "Over 45",
        "Over 50",
        "Over 60",
        "Over 70",
        "Over 80",
        "Over 90",
        "Over 100",
    ],
    "Sales growthpast 5 years": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Positive Low (0-10%)",
        "High (>25%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "Return on Investment": [
        "Any",
        "Positive (>0%)",
        "Negative (<0%)",
        "Very Positive (>25%)",
        "Very Negative (<-10%)",
        "Under -50%",
        "Under -45%",
        "Under -40%",
        "Under -35%",
        "Under -30%",
        "Under -25%",
        "Under -20%",
        "Under -15%",
        "Under -10%",
        "Under -5%",
        "Over +5%",
        "Over +10%",
        "Over +15%",
        "Over +20%",
        "Over +25%",
        "Over +30%",
        "Over +35%",
        "Over +40%",
        "Over +45%",
        "Over +50%",
    ],
    "Gross Margin": [
        "Any",
        "Positive (>0%)",
        "Negative (<0%)",
        "High (>50%)",
        "Under 90%",
        "Under 80%",
        "Under 70%",
        "Under 60%",
        "Under 50%",
        "Under 45%",
        "Under 40%",
        "Under 35%",
        "Under 30%",
        "Under 25%",
        "Under 20%",
        "Under 15%",
        "Under 10%",
        "Under 5%",
        "Under 0%",
        "Under -10%",
        "Under -20%",
        "Under -30%",
        "Under -50%",
        "Under -70%",
        "Under -100%",
        "Over 0%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
        "Over 40%",
        "Over 45%",
        "Over 50%",
        "Over 60%",
        "Over 70%",
        "Over 80%",
        "Over 90%",
    ],
    "InsiderTransactions": [
        "Any",
        "Very Negative (<20%)",
        "Negative (<0%)",
        "Positive (>0%)",
        "Very Positive (>20%)",
        "Under -90%",
        "Under 80%",
        "Under 70%",
        "Under -60%",
        "Under -50%",
        "Under -45%",
        "Under 40%",
        "Under -35%",
        "Under -30%",
        "Under -25%",
        "Under -20%",
        "Under -15%",
        "Under -10%",
        "Under -5%",
        "Over +5%",
        "Over +10%",
        "Over +15%",
        "Over +20%",
        "Over +25%",
        "Over +30%",
        "Over +35%",
        "Over +40%",
        "Over +45%",
        "Over +50%",
        "Over +60%",
        "Over +70%",
        "Over +80%",
        "Over +90%",
    ],
    "PEG": [
        "Any",
        "Low (<1)",
        "High (>2)",
        "Under 1",
        "Under 2",
        "Under 3",
        "Over 1",
        "Over 2",
        "Over 3",
    ],
    "EPS growththis year": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Positive Low (0-10%)",
        "High (>25%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "EPS growthqtr over qtr": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Positive Low (0-10%)",
        "High (>25%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "Current Ratio": [
        "Any",
        "High (>3)",
        "Low (<1)",
        "Under 1",
        "Under 0.5",
        "Over 0.5",
        "Over 1",
        "Over 1.5",
        "Over 2",
        "Over 3",
        "Over 4",
        "Over 5",
        "Over 10",
    ],
    "Operating Margin": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Very Negative (<-20%)",
        "High (>25%)",
        "Under 90%",
        "Under 80%",
        "Under 70%",
        "Under 60%",
        "Under 50%",
        "Under 45%",
        "Under 40%",
        "Under 35%",
        "Under 30%",
        "Under 25%",
        "Under 20%",
        "Under 15%",
        "Under 10%",
        "Under 5%",
        "Under 0%",
        "Under -10%",
        "Under -20%",
        "Under -30%",
        "Under -50%",
        "Under -70%",
        "Under -100%",
        "Over 0%",
        "Over 10%",
        "Under 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
        "Over 35%",
        "Over 40%",
        "Over 45%",
        "Over 50%",
        "Over 60%",
        "Over 70%",
        "Over 80%",
        "Over 90%",
    ],
    "InstitutionalOwnership": [
        "Any",
        "Low (<5%)",
        "High (>90%)",
        "Under 90%",
        "Under 80%",
        "Under 70%",
        "Under 60%",
        "Under 50%",
        "Under 40%",
        "Under 30%",
        "Under 20%",
        "Under 10%",
        "Over 10%",
        "Over 20%",
        "Over 30%",
        "Over 40%",
        "Over 50%",
        "Over 60%",
        "Over 70%",
        "Over 80%",
        "Over 90%",
    ],
    "P/S": [
        "Any",
        "Low (<1)",
        "High (>10)",
        "Under 1",
        "Under 2",
        "Under 3",
        "Under 4",
        "Under 5",
        "Under 6",
        "Under 7",
        "Under 8",
        "Under 9",
        "Under 10",
        "Over 1",
        "Over 2",
        "Over 3",
        "Over 4",
        "Over 5",
        "Over 6",
        "Over 6",
        "Over 7",
        "Over 8",
        "Over 9",
        "Over 10",
    ],
    "EPS growthnext year": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Positive Low (0-10%)",
        "High (>25%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "Sales growthqtr over qtr": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Positive Low (0-10%)",
        "High (>25%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "Quick Ratio": [
        "Any",
        "High (>3)",
        "Low (<0.5)",
        "Under 1",
        "Under 0.5",
        "Over 0.5",
        "Over 1",
        "Over 1.5",
        "Over 2",
        "Over 3",
        "Over 4",
        "Over 5",
        "Over 10",
    ],
    "Net Profit Margin": [
        "Any",
        "Positive (>0%)",
        "Negative (<0%)",
        "Very Negative (<-20%)",
        "High (>20%)",
        "Under 90%",
        "Under 80%",
        "Under 70%",
        "Under 60%",
        "Under 50%",
        "Under 45%",
        "Under 40%",
        "Under 35%",
        "Under 30%",
        "Under 25%",
        "Under 20%",
        "Under 15%",
        "Under 10%",
        "Under 5%",
        "Under 0%",
        "Under -10%",
        "Under -20%",
        "Under -30%",
        "Under -50%",
        "Under -70%",
        "Under -100%",
        "Over 0%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
        "Over 35%",
        "Over 40%",
        "Over 45%",
        "Over 50%",
        "Over 60%",
        "Over 70%",
        "Over 80%",
        "Over 90%",
    ],
    "InstitutionalTransactions": [
        "Any",
        "Very Negative (<20%)",
        "Negative (<0%)",
        "Positive (>0%)",
        "Very Positive (>20%)",
        "Under -50%",
        "Under -45%",
        "Under -40%",
        "Under -35%",
        "Under -30%",
        "Under -25%",
        "Under -20%",
        "Under -15%",
        "Under -10%",
        "Under -5%",
        "Over +5%",
        "Over +10%",
        "Over +15%",
        "Over +20%",
        "Over +25%",
        "Over +30%",
        "Over +35%",
        "Over +40%",
        "Over +45%",
        "Over +50%",
    ],
    "P/B": [
        "Any",
        "Low (<1)",
        "High (>5)",
        "Under 1",
        "Under 2",
        "Under 3",
        "Under 4",
        "Under 5",
        "Under 6",
        "Under 7",
        "Under 8",
        "Under 9",
        "Under 10",
        "Over 1",
        "Over 2",
        "Over 3",
        "Over 4",
        "Over 5",
        "Over 6",
        "Over 7",
        "Over 8",
        "Over 9",
        "Over 10",
    ],
    "EPS growthpast 5 years": [
        "Any",
        "Negative (<0%)",
        "Positive (>0%)",
        "Positive Low (0-10%)",
        "High (>25%)",
        "Under 5%",
        "Under 10%",
        "Under 15%",
        "Under 20%",
        "Under 25%",
        "Under 30%",
        "Over 5%",
        "Over 10%",
        "Over 15%",
        "Over 20%",
        "Over 25%",
        "Over 30%",
    ],
    "Return on Assets": [
        "Any",
        "Positive (>0%)",
        "Negative (<0%)",
        "Very Positive (>15%)",
        "Very Negative (<-15%)",
        "Under -50%",
        "Under -45%",
        "Under -40%",
        "Under -35%",
        "Under -30%",
        "Under -25%",
        "Under -20%",
        "Under -15%",
        "Under -10%",
        "Under -5%",
        "Over +5%",
        "Over +10%",
        "Over +15%",
        "Over +20%",
        "Over +25%",
        "Over +30%",
        "Over +35%",
        "Over +40%",
        "Over +45%",
        "Over +50%",
    ],
    "LT Debt/Equity": [
        "Any",
        "High (>0.5)",
        "Low (<0.1)",
        "Under 1",
        "Under 0.9",
        "Under 0.8",
        "Under 0.7",
        "Under 0.6",
        "Under 0.5",
        "Under 0.4",
        "Under 0.3",
        "Under 0.2",
        "Under 0.1",
        "Over 0.1",
        "Over 0.2",
        "Over 0.3",
        "Over 0.4",
        "Over.5",
        "Over 0.6",
        "Over 0.7",
        "Over 0.8",
        "Over 0.9",
        "Over 1",
    ],
    "Payout Ratio": [
        "Any",
        "None (0%)",
        "Positive (>0%)",
        "Low (<20%)",
        "High (>50%)",
        "Over 0%",
        "Over 10%",
        "Over 20%",
        "Over 30%",
        "Over 40%",
        "Over 50%",
        "Over 60%",
        "Over 70%",
        "Over 80%",
        "Over 90%",
        "Over 100%",
        "Under 10%",
        "Under 20%",
        "Under 30%",
        "Under 40%",
        "Under 50%",
        "Under 60%",
        "Under 70%",
        "Under 80%",
        "Under 90%",
        "Under 100%",
    ],
    "Performance": [
        "Any",
        "Today Up",
        "Today Down",
        "Today -15%",
        "Today -10%",
        "Today -5%",
        "Today +5%",
        "Today +10%",
        "Today +15%",
        "Week -30%",
        "Week -20%",
        "Week -10%",
        "Week Down",
        "Week Up",
        "Week +10%",
        "Week +20%",
        "Week +30%",
        "Month -50%",
        "Month -30%",
        "Month -20%",
        "Month -10%",
        "Month Down",
        "Month Up",
        "Month +10%",
        "Month +20%",
        "Month +30%",
        "Month +50%",
        "Quarter -50%",
        "Quarter -30%",
        "Quarter -20%",
        "Quarter -10%",
        "Quarter Down",
        "Quarter Up",
        "Quarter +10%",
        "Quarter +20%",
        "Quarter +30%",
        "Quarter +50%",
        "Half -75%",
        "Half -50%",
        "Half -30%",
        "Half -20%",
        "Half -10%",
        "Half Down",
        "Half Up",
        "Half +10%",
        "Half +20%",
        "Half +30%",
        "Half +50%",
        "Half +100%",
        "Year -75%",
        "Year -50%",
        "Year -30%",
        "Year -20%",
        "Year -10%",
        "Year Down",
        "Year Up",
        "Year +10%",
        "Year +20%",
        "Year +30%",
        "Year +50%",
        "Year +100%",
        "Year +200%",
        "Year +300%",
        "Year +500%",
        "YTD -75%",
        "YTD -50%",
        "YTD -30%",
        "YTD -20%",
        "YTD -10%",
        "YTD -5%",
        "YTD Down",
        "YTD Up",
        "YTD +5%",
        "YTD +10%",
        "YTD +20%",
        "YTD +30",
        "YTD +50%",
        "YTD +100%",
    ],
    "20-Day Simple Moving Average": [
        "Any",
        "Price below SMA20",
        "Price 10% below SMA20",
        "Price 20% below SMA20",
        "Price 30% below SMA20",
        "Price 40% below SMA20",
        "Price 50% below SMA20",
        "Price above SMA20",
        "Price 10% above SMA20",
        "Price 20% above SMA20",
        "Price 30% above SMA20",
        "Price 40% above SMA20",
        "Price 50% above SMA20",
        "Price crossed SMA20",
        "Price crossed SMA20 above",
        "Price crossed SMA20 below",
        "SMA20 crossed SMA50",
        "SMA20 crossed SMA50 above",
        "SMA20 crossed SMA50 below",
        "SMA20 cross SMA200",
        "SMA20 crossed SMA200 below",
        "SMA20 crossed SMA200 above",
        "SMA20 above SMA50",
        "SMA20 below SMA50",
        "SMA20 above SMA200",
        "SMA20 below SMA200",
    ],
    "20-Day High/Low": [
        "Any",
        "New High",
        "New Low",
        "5% or more below High",
        "10% or more below High",
        "15% or more below High",
        "20% or more below High",
        "30% or more below High",
        "40% or more below High",
        "50% or more below High",
        "0-3% below High",
        "0-5% below High",
        "0-10% below High",
        "5% or more above Low",
        "10% or more above Low",
        "15% or more above Low",
        "20% or more above Low",
        "30% or more above Low",
        "40% or more above Low",
        "50% or more above Low",
        "0-3% above Low",
        "0-5% above Low",
        "0-10% above Low",
    ],
    "Beta": [
        "Any",
        "Under 0",
        "Under 0.5",
        "Under 1",
        "Under 1.5",
        "Under 2",
        "Over 0",
        "Over 0.5",
        "Over 1",
        "Over 1.5",
        "Over 2",
        "Over 2.5",
        "Over 3",
        "Over 4",
        "0 to 0.5",
        "0 to 1",
        "0.5 to 1",
        "0.5 to 1.5",
        "1 to 1.5",
        "1 to 2",
    ],
    "Performance 2": [
        "Any",
        "Today Up",
        "Today Down",
        "Today -15%",
        "Today -10%",
        "Today -5%",
        "Today +5%",
        "Today +10%",
        "Today +15%",
        "Week -30%",
        "Week -20%",
        "Week -10%",
        "Week Down",
        "Week Up",
        "Week +10%",
        "Week +20%",
        "Week +30%",
        "Month -50%",
        "Month -30%",
        "Month -20%",
        "Month -10%",
        "Month Down",
        "Month Up",
        "Month +10%",
        "Month +20%",
        "Month +30%",
        "Month +50%",
        "Quarter -50%",
        "Quarter -30%",
        "Quarter -20%",
        "Quarter -10%",
        "Quarter Down",
        "Quarter Up",
        "Quarter +10%",
        "Quarter +20%",
        "Quarter +30%",
        "Quarter +50%",
        "Half -75%",
        "Half -50%",
        "Half -30%",
        "Half -20%",
        "Half -10%",
        "Half Down",
        "Half Up",
        "Half +10%",
        "Half +20%",
        "Half +30%",
        "Half +50%",
        "Half +100%",
        "Year -75%",
        "Year -50%",
        "Year -30%",
        "Year -20%",
        "Year -10%",
        "Year Down",
        "Year Up",
        "Year +10%",
        "Year +20%",
        "Year +30%",
        "Year +50%",
        "Year +100%",
        "Year +200%",
        "Year +300%",
        "Year +500%",
        "YTD -75%",
        "YTD -50%",
        "YTD -30%",
        "YTD -20%",
        "YTD -10%",
        "YTD -5%",
        "YTD Down",
        "YTD Up",
        "YTD +5%",
        "YTD +10%",
        "YTD +20%",
        "YTD +30",
        "YTD +50%",
        "YTD +100%",
    ],
    "50-Day Simple Moving Average": [
        "Any",
        "Price below SMA50",
        "Price 10% below SMA50",
        "Price 20% below SMA50",
        "Price 30% below SMA50",
        "Price 40% below SMA50",
        "Price 50% below SMA50",
        "Price above SMA50",
        "Price 10% above SMA50",
        "Price 20% above SMA50",
        "Price 30% above SMA50",
        "Price 40% above SMA50",
        "Price 50% above SMA50",
        "Price crossed SMA50",
        "Price crossed SMA50 above",
        "Price crossed SMA50 below",
        "SMA50 crossed SMA20",
        "SMA50 crossed SMA20 above",
        "SMA50 crossed SMA20 below",
        "SMA50 cross SMA200",
        "SMA50 crossed SMA200 below",
        "SMA50 crossed SMA200 above",
        "SMA50 above SMA20",
        "SMA50 below SMA20",
        "SMA50 above SMA200",
        "SMA50 below SMA200",
    ],
    "50-Day High/Low": [
        "Any",
        "New High",
        "New Low",
        "5% or more below High",
        "10% or more below High",
        "15% or more below High",
        "20% or more below High",
        "30% or more below High",
        "40% or more below High",
        "50% or more below High",
        "0-3% below High",
        "0-5% below High",
        "0-10% below High",
        "5% or more above Low",
        "10% or more above Low",
        "15% or more above Low",
        "20% or more above Low",
        "30% or more above Low",
        "40% or more above Low",
        "50% or more above Low",
        "0-3% above Low",
        "0-5% above Low",
        "0-10% above Low",
    ],
    "Average True Range": [
        "Any",
        "Over 0.25",
        "Over 0.5",
        "Over 0.75",
        "Over 1",
        "Over 1.5",
        "Over 2. Over 2.5",
        "Over 3",
        "Over 3.5",
        "Over 4",
        "Over 4.5",
        "Over 5",
        "Under 0.25",
        "Under 0.5",
        "Under 0.75",
        "Under 1",
        "Under 1.5",
        "Under 2",
        "Under 2.5",
        "Under 3",
        "Under 3.5",
        "Under 4",
        "Under 4.5",
        "Under 5",
    ],
    "Volatility": [
        "Any",
        "Week - Over 3%",
        "Week - Over 4%",
        "Week - Over 5%",
        "Week - 6%",
        "Week - 7%",
        "Week - 8%",
        "Week - 9%",
        "Week - 10%",
        "Week - 12%",
        "Week - 15%",
        "Month - 2%",
        "Month - 3%",
        "Month - 4%",
        "Month 5%",
        "Month 5%",
        "Month 6%",
        "Month 7%",
        "Month 8%",
        "Month 9%",
        "Month 10%",
        "Month 12%",
        "Month 15%",
    ],
    "200-Day Simple Moving Average": [
        "Any",
        "Price below SMA200",
        "Price 10% below SMA200",
        "Price 20% below SMA200",
        "Price 30% below SMA200",
        "Price 40% below SMA200",
        "Price 50% below SMA200",
        "Price above SMA200",
        "Price 10% above SMA200",
        "Price 20% above SMA200",
        "Price 30% above SMA200",
        "Price 40% above SMA200",
        "Price 50% above SMA200",
        "Price crossed SMA200",
        "Price crossed SMA200 above",
        "Price crossed SMA200 below",
        "SMA200 crossed SMA20",
        "SMA20 crossed SMA20 above",
        "SMA20 crossed SMA20 below",
        "SMA200 cross SMA50",
        "SMA200 crossed SMA50 below",
        "SMA200 crossed SMA50 above",
        "SMA200 above SMA20",
        "SMA200 below SMA20",
        "SMA200 above SMA50",
        "SMA200 below SMA50",
    ],
    "52-Week High/Low": [
        "Any",
        "New High",
        "New Low",
        "5% or more below High",
        "10% or more below High",
        "15% or more below High",
        "20% or more below High",
        "30% or more below High",
        "40% or more below High",
        "50% or more below High",
        "0-3% below High",
        "0-5% below High",
        "0-10% below High",
        "5% or more above Low",
        "10% or more above Low",
        "15% or more above Low",
        "20% or more above Low",
        "30% or more above Low",
        "40% or more above Low",
        "50% or more above Low",
        "0-3% above Low",
        "0-5% above Low",
        "0-10% above Low",
    ],
    "RSI (14)": [
        "Any",
        "Overbought (90)",
        "Overbought (80)",
        "Overbought (70)",
        "Overbought (6)",
        "Oversold (40)",
        "Oversold (30)",
        "Oversold (20)",
        "Oversold (10)",
        "Not Overbought (<60)",
        "Not Overbought (<50)",
        "Not Oversold (>50)",
        "Not Oversold (>40)",
    ],
    "Change": [
        "Any",
        "Up",
        "Up 1%",
        "Up 2%",
        "Up 3%",
        "Up 4%",
        "Up 5%",
        "Up 6%",
        "Up 7%",
        "Up 8%",
        "Up 9%",
        "Up 10%",
        "Up 15%",
        "Up 20%",
        "Down",
        "Down 1%",
        "Down 2%",
        "Down 3%",
        "Down 4%",
        "Down 5%",
        "Down 6%",
        "Down 7%",
        "Down 8%",
        "Down 9%",
        "Down 10%",
        "Down 15%",
        "Down 20%",
    ],
    "Pattern": [
        "Any",
        "Horizontal S/R",
        "Horizontal S/R (Strong)",
        "TL Resistance",
        "TL Resistance (Strong)",
        "TL Support",
        "TL Support (Strong)",
        "Wedge Up",
        "Wedge Up (Strong)",
        "Wedge Down",
        "Wedge Down (Strong)",
        "Triangle Ascending",
        "Triangle Ascending (Strong)",
        "Triangle Descending",
        "Triangle Descending (Strong)",
        "Wedge",
        "Wedge (Strong)",
        "Channel Up",
        "Channel Up (Strong)",
        "Channel Down",
        "Channel Down (Strong)",
        "Channel",
        "Channel (Strong)",
        "Double Top",
        "Double Bottom",
        "Multiple Top",
        "Multiple Bottom",
        "Head & Shoulders",
        "Head & Shoulders Inverse",
    ],
    "Gap": [
        "Any",
        "Up",
        "Up 1%",
        "Up 2%",
        "Up 3%",
        "Up 4%",
        "Up 5%",
        "Up 6%",
        "Up 7%",
        "Up 8%",
        "Up 9%",
        "Up 10%",
        "Up 15%",
        "Up 20%",
        "Down",
        "Down 1%",
        "Down 2%",
        "Down 3%",
        "Down 4%",
        "Down 5%",
        "Down 6%",
        "Down 7%",
        "Down 8%",
        "Down 9%",
        "Down 10%",
        "Down 15%",
        "Down 20%",
    ],
    "Change from Open": [
        "Any",
        "Up",
        "Up 1%",
        "Up 2%",
        "Up 3%",
        "Up 4%",
        "Up 5%",
        "Up 6%",
        "Up 7%",
        "Up 8%",
        "Up 9%",
        "Up 10%",
        "Up 15%",
        "Up 20%",
        "Down",
        "Down 1%",
        "Down 2%",
        "Down 3%",
        "Down 4%",
        "Down 5%",
        "Down 6%",
        "Down 7%",
        "Down 8%",
        "Down 9%",
        "Down 10%",
        "Down 15%",
        "Down 20%",
    ],
    "Candlestick": [
        "Any",
        "Long Lower Shadow",
        "Long Upper Shadow",
        "Hammer",
        "Inverted Hammer",
        "Spinning Top White",
        "Spinning Top Black",
        "Doji",
        "Dragonfly Doji",
        "Gravestone Doji",
        "Marubozu White",
        "Marubozu Black",
    ],
}

SIGNALS = sorted(list(d_signals))

SIGNALS_DESC_STR = ""
for k, v in sorted(d_signals_desc.items()):
    SIGNALS_DESC_STR += f"\n        {k}: {v}"


SECTOR_MAP = {
    "all": "Any",
    "energy": "Energy",
    "materials": "Basic Materials",
    "industrials": "Industrials",
    "consumer_cyclical": "Consumer Cyclical",
    "consumer_defensive": "Consumer Defensive",
    "financial": "Financial",
    "healthcare": "Healthcare",
    "technology": "Technology",
    "communication_services": "Communication Services",
    "utilities": "Utilities",
    "real_estate": "Real Estate",
}

Sectors = Literal[
    "all",
    "energy",
    "materials",
    "industrials",
    "consumer_cyclical",
    "consumer_defensive",
    "financial",
    "healthcare",
    "technology",
    "communication_services",
    "utilities",
    "real_estate",
]

EXCHANGE_MAP = {
    "all": "Any",
    "amex": "AMEX",
    "nasdaq": "NASDAQ",
    "nyse": "NYSE",
}

INDEX_MAP = {
    "all": "Any",
    "dow": "DJIA",
    "sp500": "S&P 500",
    "nasdaq": "NASDAQ 100",
    "russell": "RUSSELL 2000",
}


def get_preset_choices(user_data_path) -> Dict:
    """Get a combined map of default and user screener presets."""
    # pylint: disable=import-outside-toplevel
    import shutil
    from pathlib import Path
    from warnings import warn

    PRESETS_PATH = Path(user_data_path) / "presets" / "finviz"
    PRESETS_PATH_DEFAULT = Path(__file__).parent.resolve() / "presets"
    preset_choices: Dict = {}

    if PRESETS_PATH_DEFAULT.exists():
        preset_choices.update(
            {
                filepath.name.replace(".ini", ""): filepath
                for filepath in PRESETS_PATH_DEFAULT.iterdir()
                if filepath.suffix == ".ini"
            }
        )

    try:
        # Create the user presets directory if it doesn't exist.
        if not PRESETS_PATH.exists():
            PRESETS_PATH.mkdir(parents=True, exist_ok=True)

        # Copy any missing default presets to the user presets directory.
        for filepath in PRESETS_PATH_DEFAULT.iterdir():
            if filepath.suffix == ".ini":
                target_path = PRESETS_PATH / filepath.name
                if not target_path.exists():
                    shutil.copy(filepath, target_path)

        # Override any default paths to the user path, if they exist.
        if PRESETS_PATH.exists():
            preset_choices.update(
                {
                    filepath.name.replace(".ini", ""): filepath
                    for filepath in PRESETS_PATH.iterdir()
                    if filepath.suffix == ".ini"
                }
            )
    except Exception as e:
        warn(f"Error loading user presets: {e}")

    preset_choices = {
        k: v for k, v in sorted(preset_choices.items()) if k != "screener_template"
    }

    return preset_choices



"""Finviz Price Performance Model."""

# pylint: disable=unused-argument
from typing import Any, Dict, List, Optional
from warnings import warn

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.recent_performance import (
    RecentPerformanceData,
    RecentPerformanceQueryParams,
)
from openbb_core.provider.utils.errors import EmptyDataError
from pydantic import Field


class FinvizPricePerformanceQueryParams(RecentPerformanceQueryParams):
    """
    Finviz Price Performance Query.

    Source: https://finviz.com/screener.ashx
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FinvizPricePerformanceData(RecentPerformanceData):
    """Finviz Price Performance Data."""

    __alias_dict__ = {
        "symbol": "Ticker",
        "one_week": "Perf Week",
        "one_month": "Perf Month",
        "three_month": "Perf Quart",
        "six_month": "Perf Half",
        "one_year": "Perf Year",
        "ytd": "Perf YTD",
        "volatility_week": "Volatility W",
        "volatility_month": "Volatility M",
        "price": "Price",
        "one_day": "Change",
        "volume": "Volume",
        "average_volume": "Avg Volume",
        "relative_volume": "Rel Volume",
        "analyst_score": "Recom",
    }

    volatility_week: Optional[float] = Field(
        default=None,
        description="One-week realized volatility, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    volatility_month: Optional[float] = Field(
        default=None,
        description="One-month realized volatility, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    price: Optional[float] = Field(
        default=None,
        description="Last Price.",
    )
    volume: Optional[float] = Field(
        default=None,
        description="Current volume.",
    )
    average_volume: Optional[float] = Field(
        default=None,
        description="Average daily volume.",
    )
    relative_volume: Optional[float] = Field(
        default=None,
        description="Relative volume as a ratio of current volume to average volume.",
    )
    analyst_recommendation: Optional[float] = Field(
        default=None,
        description="The analyst consensus, on a scale of 1-5 where 1 is a buy and 5 is a sell.",
    )
    symbol: Optional[str] = Field(
        default=None,
        description="The ticker symbol.",
    )


class FinvizPricePerformanceFetcher(
    Fetcher[FinvizPricePerformanceQueryParams, List[FinvizPricePerformanceData]]
):
    """Finviz Price Performance Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinvizPricePerformanceQueryParams:
        """Transform the query params."""
        return FinvizPricePerformanceQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FinvizPricePerformanceQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the raw data from Finviz."""
        # pylint: disable=import-outside-toplevel
        from finvizfinance import util
        from finvizfinance.screener import performance
        from openbb_core.provider.utils.helpers import get_requests_session

        util.session = get_requests_session()
        screen = performance.Performance()
        screen.set_filter(ticker=query.symbol)
        try:
            screen_df = screen.screener_view(verbose=0)
            if screen_df is None:
                raise EmptyDataError()
            screen_df.columns = screen_df.columns.str.strip()  # type: ignore
            screen_df = screen_df.fillna("N/A").replace("N/A", None)  # type: ignore
        except Exception as e:
            raise e from e

        symbols = query.symbol.split(",")

        # Check for missing symbols and warn of the missing symbols.
        for symbol in symbols:
            if symbol not in screen_df["Ticker"].tolist():
                warn(f"Symbol Error: {symbol} was not found.")

        return sorted(
            screen_df.to_dict(orient="records"),
            key=(lambda item: (symbols.index(item.get("Ticker", len(symbols))))),
        )

    @staticmethod
    def transform_data(
        query: FinvizPricePerformanceQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[FinvizPricePerformanceData]:
        """Transform the raw data."""
        return [FinvizPricePerformanceData.model_validate(d) for d in data]



"""Finviz provider module."""

from openbb_core.provider.abstract.provider import Provider
from openbb_finviz.models.compare_groups import FinvizCompareGroupsFetcher
from openbb_finviz.models.equity_profile import FinvizEquityProfileFetcher
from openbb_finviz.models.equity_screener import FinvizEquityScreenerFetcher
from openbb_finviz.models.key_metrics import FinvizKeyMetricsFetcher
from openbb_finviz.models.price_performance import FinvizPricePerformanceFetcher
from openbb_finviz.models.price_target import FinvizPriceTargetFetcher

finviz_provider = Provider(
    name="finviz",
    website="https://finviz.com",
    description="Unofficial Finviz API - https://github.com/lit26/finvizfinance/releases",
    credentials=None,
    fetcher_dict={
        "CompareGroups": FinvizCompareGroupsFetcher,
        "EtfPricePerformance": FinvizPricePerformanceFetcher,
        "EquityInfo": FinvizEquityProfileFetcher,
        "EquityScreener": FinvizEquityScreenerFetcher,
        "KeyMetrics": FinvizKeyMetricsFetcher,
        "PricePerformance": FinvizPricePerformanceFetcher,
        "PriceTarget": FinvizPriceTargetFetcher,
    },
    repr_name="FinViz",
)



"""Finviz Key Metrics Model."""

# pylint: disable=unused-argument

from typing import Any, Dict, List, Optional
from warnings import warn

from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.key_metrics import (
    KeyMetricsData,
    KeyMetricsQueryParams,
)
from pydantic import Field


class FinvizKeyMetricsQueryParams(KeyMetricsQueryParams):
    """
    Finviz Key Metrics Query.

    Source: https://finviz.com/screener.ashx
    """

    __json_schema_extra__ = {"symbol": {"multiple_items_allowed": True}}


class FinvizKeyMetricsData(KeyMetricsData):
    """Finviz Key Metrics Data."""

    foward_pe: Optional[float] = Field(
        default=None, description="Forward price-to-earnings ratio (forward P/E)"
    )
    eps: Optional[float] = Field(default=None, description="Earnings per share (EPS)")
    price_to_sales: Optional[float] = Field(
        default=None, description="Price-to-sales ratio (P/S)"
    )
    price_to_book: Optional[float] = Field(
        default=None, description="Price-to-book ratio (P/B)"
    )
    book_value_per_share: Optional[float] = Field(
        default=None, description="Book value per share (Book/sh)"
    )
    price_to_cash: Optional[float] = Field(
        default=None, description="Price-to-cash ratio (P/C)"
    )
    cash_per_share: Optional[float] = Field(
        default=None, description="Cash per share (Cash/sh)"
    )
    price_to_free_cash_flow: Optional[float] = Field(
        default=None, description="Price-to-free cash flow ratio (P/FCF)"
    )
    debt_to_equity: Optional[float] = Field(
        default=None, description="Debt-to-equity ratio (Debt/Eq)"
    )
    long_term_debt_to_equity: Optional[float] = Field(
        default=None, description="Long-term debt-to-equity ratio (LT Debt/Eq)"
    )
    quick_ratio: Optional[float] = Field(default=None, description="Quick ratio")
    current_ratio: Optional[float] = Field(default=None, description="Current ratio")
    gross_margin: Optional[float] = Field(
        default=None,
        description="Gross margin, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    profit_margin: Optional[float] = Field(
        default=None,
        description="Profit margin, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    operating_margin: Optional[float] = Field(
        default=None,
        description="Operating margin, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_assets: Optional[float] = Field(
        default=None,
        description="Return on assets (ROA), as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_investment: Optional[float] = Field(
        default=None,
        description="Return on investment (ROI), as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    return_on_equity: Optional[float] = Field(
        default=None,
        description="Return on equity (ROE), as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    payout_ratio: Optional[float] = Field(
        default=None,
        description="Payout ratio, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )
    dividend_yield: Optional[float] = Field(
        default=None,
        description="Dividend yield, as a normalized percent.",
        json_schema_extra={"x-unit_measurement": "percent", "x-frontend_multiply": 100},
    )


class FinvizKeyMetricsFetcher(
    Fetcher[FinvizKeyMetricsQueryParams, List[FinvizKeyMetricsData]]
):
    """Finviz Key Metrics Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> FinvizKeyMetricsQueryParams:
        """Transform the query params."""
        return FinvizKeyMetricsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: FinvizKeyMetricsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the raw data from Finviz."""
        # pylint: disable=import-outside-toplevel
        from finvizfinance import util
        from finvizfinance.quote import finvizfinance
        from openbb_core.app.model.abstract.error import OpenBBError
        from openbb_core.provider.utils.errors import EmptyDataError
        from openbb_core.provider.utils.helpers import get_requests_session

        results: List = []
        messages: List = []
        util.session = get_requests_session()

        def get_one(symbol) -> Dict:
            """Get the data for one symbol."""
            result: Dict = {}
            try:
                data = finvizfinance(symbol)
                fundament = data.ticker_fundament()
                mkt_cap = (
                    fundament.get("Market Cap", None)
                    if fundament.get("Market Cap", "-") != "-"
                    else None
                )
                if mkt_cap:
                    mkt_cap = float(
                        str(mkt_cap)
                        .replace("B", "e+9")
                        .replace("M", "e+6")
                        .replace("K", "e+3")
                    )
            except Exception as e:  # pylint: disable=W0718
                warn(f"Failed to get data for {symbol} -> {e}")
                return result
            result.update(
                {
                    "symbol": symbol,
                    "market_cap": int(mkt_cap) if mkt_cap is not None else None,
                    "pe_ratio": (
                        fundament.get("P/E", None)
                        if fundament.get("P/E", "-") != "-"
                        else None
                    ),
                    "eps": (
                        fundament.get("EPS (ttm)", None)
                        if fundament.get("EPS (ttm)", "-") != "-"
                        else None
                    ),
                    "foward_pe": (
                        fundament.get("Forward P/E", None)
                        if fundament.get("Forward P/E", "-") != "-"
                        else None
                    ),
                    "price_to_sales": (
                        fundament.get("P/S", None)
                        if fundament.get("P/S", "-") != "-"
                        else None
                    ),
                    "price_to_book": (
                        fundament.get("P/B", None)
                        if fundament.get("P/B", "-") != "-"
                        else None
                    ),
                    "book_value_per_share": (
                        fundament.get("Book/sh", None)
                        if fundament.get("Book/sh", "-") != "-"
                        else None
                    ),
                    "price_to_cash": (
                        fundament.get("P/C", None)
                        if fundament.get("P/C", "-") != "-"
                        else None
                    ),
                    "cash_per_share": (
                        fundament.get("Cash/sh", None)
                        if fundament.get("Cash/sh", "-") != "-"
                        else None
                    ),
                    "price_to_free_cash_flow": (
                        fundament.get("P/FCF", None)
                        if fundament.get("P/FCF", "-") != "-"
                        else None
                    ),
                    "debt_to_equity": (
                        fundament.get("Debt/Eq", None)
                        if fundament.get("Debt/Eq", "-") != "-"
                        else None
                    ),
                    "long_term_debt_to_equity": (
                        fundament.get("LT Debt/Eq", None)
                        if fundament.get("LT Debt/Eq", "-") != "-"
                        else None
                    ),
                    "quick_ratio": (
                        fundament.get("Quick Ratio", None)
                        if fundament.get("Quick Ratio", "-") != "-"
                        else None
                    ),
                    "current_ratio": (
                        fundament.get("Current Ratio", None)
                        if fundament.get("Current Ratio", "-") != "-"
                        else None
                    ),
                    "gross_margin": (
                        float(str(fundament.get("Gross Margin", None)).replace("%", ""))
                        / 100
                        if fundament.get("Gross Margin", "-") != "-"
                        else None
                    ),
                    "profit_margin": (
                        float(
                            str(fundament.get("Profit Margin", None)).replace("%", "")
                        )
                        / 100
                        if fundament.get("Profit Margin", "-") != "-"
                        else None
                    ),
                    "operating_margin": (
                        float(str(fundament.get("Oper. Margin", None)).replace("%", ""))
                        / 100
                        if fundament.get("Oper. Margin", "-") != "-"
                        else None
                    ),
                    "return_on_assets": (
                        float(str(fundament.get("ROA", None)).replace("%", "")) / 100
                        if fundament.get("ROA", "-") != "-"
                        else None
                    ),
                    "return_on_investment": (
                        float(str(fundament.get("ROI", None)).replace("%", "")) / 100
                        if fundament.get("ROI", "-") != "-"
                        else None
                    ),
                    "return_on_equity": (
                        float(str(fundament.get("ROE", None)).replace("%", "")) / 100
                        if fundament.get("ROE", "-") != "-"
                        else None
                    ),
                    "payout_ratio": (
                        float(str(fundament.get("Payout", None)).replace("%", "")) / 100
                        if fundament.get("Payout", "-") != "-"
                        else None
                    ),
                    "dividend_yield": (
                        float(str(fundament.get("Dividend %", None)).replace("%", ""))
                        / 100
                        if fundament.get("Dividend %", "-") != "-"
                        else None
                    ),
                }
            )

            return result

        symbols = query.symbol.split(",")

        for symbol in symbols:
            result = get_one(symbol)
            if result is not None and result:
                results.append(result)

        if not results and messages:
            raise OpenBBError("\n".join(messages))

        if not results and not messages:
            raise EmptyDataError("No data was returned for any symbol")

        if results and messages:
            for message in messages:
                warn(message)

        return results

    @staticmethod
    def transform_data(
        query: FinvizKeyMetricsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[FinvizKeyMetricsData]:
        """Transform and validate the raw data."""
        return [FinvizKeyMetricsData.model_validate(d) for d in data]
